# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
import csv
from collections import OrderedDict
import numpy as np
from MindsporeTrainer.utils.metrics import *
from mindspore.communication import get_rank, get_group_size

from MindsporeTrainer.modeling.layers import FakeHead


__all__ = ['EvalData', 'Task', 'TransformerTask']

class EvalData:
    def __init__(self, name, examples, metrics_fn=None, predict_fn=None, ignore_metric=False, critial_metrics=None):
        def accuracy_fn(logits, labels):
            return OrderedDict(accuracy= metric_accuracy(logits, labels))

        def default_pred_fn(logits, output_dir, name, prefix):
            output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
            preds = np.argmax(logits, axis=-1)
            with open(output, 'w', encoding='utf-8') as fs:
                fs.write('index\tpredictions\n')
                for i,p in enumerate(preds):
                    fs.write('{}\t{}\n'.format(i, p))
        self.name = name
        self.data = examples
        self.ignore_metric = ignore_metric
        self.critial_metrics = critial_metrics
        self.metrics_fn = metrics_fn if metrics_fn is not None else accuracy_fn
        self.predict_fn = predict_fn if predict_fn is not None else default_pred_fn

    def __repr__(self):
        return f'{self.name}, {type(self.data)}: {len(self.data)}, {self.predict_fn}, {self.metrics_fn}'

class Task():
    _meta={}

    def __init__(self, args, **kwargs):
        self.args = args
        self.data_dir = args.data_dir
        self.num_train_epochs = self.args.num_train_epochs
        self.data_workers = self.args.data_workers
        self.train_batch_size = self.args.train_batch_size
        self.eval_batch_size = self.args.eval_batch_size
        self.predict_batch_size = self.args.predict_batch_size
        self.output_dir = self.args.output_dir
        if args.distribute:
            self.rank_id = get_rank()
            self.rank_size = get_group_size()
        else:
            self.rank_id = 0
            self.rank_size = 1

    def eval_data(self, **kwargs):
        """
        Get eval dataset object.
        """
        return None

    def train_data(self, **kwargs):
        """
        Get train dataset object.
        """
        return None

    def test_data(self, **kwargs):
        return None

    def get_labels(self):
        """Gets the list of labels for this data set."""
        return None

    def label2id(self, labelstr):
        label_dict = {l:i for i,l in enumerate(self.get_labels())}
        return label_dict[labelstr] if labelstr in label_dict else -1

    def get_train_fn(self, *args, **kwargs):
        return None

    def get_eval_fn(self, *args, **kwargs):
        """
        Get the evaluation function
        """
        return None

    def get_eval_head(self, *args, **kwargs):
        """
        Get the evaluate head, the head replace loss function head when evaluation process
        """
        return FakeHead()

    def get_pred_fn(self, *args, **kwargs):
        """
        Get the predict function
        """
        return None

    def get_loss(self, *args, **kwargs):
        """
        Get the loss function
        """
        return None

    def get_opt_fn(self, *args, **kwargs):
        """
        Get a function wich return the opimizer
        """
        def get_optimizer(*args, **kwargs):
            pass
        return get_optimizer

    def get_metrics(self):
        """Calcuate metrics based on prediction results"""
        return None

    def get_predict_fn(self):
        """Calcuate metrics based on prediction results"""
        def predict_fn(logits, output_dir, name, prefix):
            output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
            preds = np.argmax(logits, axis=-1)
            preds = preds.flatten()
            labels = self.get_labels()
            with open(output, 'w', encoding='utf-8') as fs:
                fs.write('index\tpredictions\n')
                for i,p in enumerate(preds):
                    fs.write('{}\t{}\n'.format(i, labels[p]))

        return predict_fn

    @classmethod
    def _read_tsv(cls, input_file, quotechar=None):
        """Reads a tab separated value file."""
        with open(input_file, "r", encoding='utf-8') as f:
            reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
            lines = []
            for line in reader:
                lines.append(line)
            return lines

    def get_feature_fn(self, max_seq_len = 512, mask_gen = None, label_type='int', training=False):
        # tokenizer = self.tokenizer
        # def _example_to_feature(example, rng=None, ext_params=None, **kwargs):
        #     return example_to_feature(tokenizer, example, max_seq_len = max_seq_len, \
        #         rng = rng, mask_generator = mask_gen, ext_params = ext_params, label_type=label_type, **kwargs)
        # return _example_to_feature
        return None

    def get_model_class_fn(self):
        # def partial_class(*wargs, **kwargs):
        #     model, tokenizer, model_config = NNModule.load_model(*wargs, self.model_class, **kwargs)
        #     self.tokenizer = tokenizer
        #     self.model_config = model_config
        #     if 'ForMaskedLM' in model.__class__.__name__:
        #         model = self.custom_model_class(model_config, model, num_labels=kwargs['num_labels'])
        #     return model
        # return partial_class
        return None
    
    def get_model(self):
        """
        Get a model instance
        """
        raise NotImplementedError('method not implemented yet.')

    def get_callbacks(self, *args, **kwargs):
        return None

    @classmethod
    def add_arguments(cls, parser):
        """Add task specific arguments
            e.g. parser.add_argument('--data_dir', type=str, help='The path of data directory.')
        """
        # parser.add_argument('--task_example_arg', type=str, default=None, help='An example task specific argument')

        return parser


class TransformerTask(Task):
    _meta={}
    max_seq_len = 512
    model_config = ''
    vocab_type = ''
    vocab_path = ''
    def __init__(self, args, **kwargs):
        self.args = args
        super().__init__(args, **kwargs)


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import argparse
from ..utils import boolean_string

__all__ = ['get_args']
def get_args():
  parser=argparse.ArgumentParser(add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  group = parser.add_argument_group(title='Optimizer', description='Parameters for the distributed optimizer')
  group.add_argument('--fp16',
            default=False,
            type=boolean_string,
            help="Whether to use 16-bit float precision instead of 32-bit")

  group.add_argument('--loss_scale',
            type=float, default=16384,
            help='Loss scaling, positive power of 2 values can improve fp16 convergence.')

  group.add_argument('--scale_steps',
            type=int, default=250,
            help='The steps to wait to increase the loss scale.')

  group.add_argument('--lookahead_k',
            default=-1,
            type=int,
            help="lookahead k parameter")

  group.add_argument('--lookahead_alpha',
            default=0.5,
            type=float,
            help="lookahead alpha parameter")

  group.add_argument('--with_radam',
            default=False,
            type=boolean_string,
            help="whether to use RAdam")

  group.add_argument('--with_lamb',
            default=False,
            type=boolean_string,
            help="whether to use RAdam")

  group.add_argument('--opt_type',
            type=str.lower,
            default='adam',
            choices=['adam', 'admax'],
            help="The optimizer to be used.")

  group.add_argument("--warmup",
            default=0.1,
            type=float,
            help="Proportion of training to perform linear learning rate warmup for. "
              "E.g., 0.1 = 10%% of training.")

  group.add_argument("--lr_schedule_ends",
            default=0,
            type=float,
            help="The ended learning rate scale for learning rate scheduling")

  group.add_argument("--lr_schedule",
            default='warmup_linear',
            type=str,
            help="The learning rate scheduler used for traning. " +
              "E.g. warmup_linear, warmup_linear_shift, warmup_cosine, warmup_constant. Default, warmup_linear")

  group.add_argument("--max_grad_norm",
            default=1,
            type=float,
            help="The clip threshold of global gradient norm")

  group.add_argument("--learning_rate",
            default=5e-5,
            type=float,
            help="The initial learning rate for Adam.")

  group.add_argument("--epsilon",
            default=1e-6,
            type=float,
            help="epsilon setting for Adam.")

  group.add_argument("--adam_beta1",
            default=0.9,
            type=float,
            help="The beta1 parameter for Adam.")

  group.add_argument("--adam_beta2",
            default=0.999,
            type=float,
            help="The beta2 parameter for Adam.")

  group.add_argument('--weight_decay',
            type=float,
            default=0.01,
            help="The weight decay rate")

  group.add_argument('--resume_opt_path',
            type=str.lower,
            default='',
            help="The optimizer to be resumed.")

  return parser




# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Parallel Loss for the Parallel Training
This is an experimental interface that is subject to change or deletion.
"""
from mindspore.common.tensor import Tensor
import mindspore.common.dtype as mstype
from mindspore.ops import operations as P
from mindspore.ops import functional as F
from mindspore.nn import Cell
from mindspore.nn.loss.loss import _check_is_tensor
from mindspore.parallel._utils import _get_parallel_mode, _is_sharding_propagation
from mindspore.context import ParallelMode
from .layers import _check_input_dtype, _check_input_shape
from .op_parallel_config import default_dpmp_config, OpParallelConfig

__all__ = ["CrossEntropyLoss"]


class CrossEntropyLoss(Cell):
    """
    Calculate the cross entropy loss.

    Args:
        parallel_config (OpParallelConfig): The parallel configure. Default `default_dpmp_config`,
            an instance of `OpParallelConfig` with default args.

    Inputs:
        - **logits** (Tensor) - Tensor of shape (N, C). Data type must be float16 or float32. The output logits of
          the backbone.

        - **labels** (Tensor) - Tensor of shape (N, ). The ground truth label of the sample.

        - **input_mask** (Tensor) - Tensor of shape (N, ). input_mask indicates whether there are padded inputs and for
          padded inputs it will not be counted into loss.

    Outputs:
        Tensor. The corresponding cross entropy loss

    Examples:
        >>> import numpy as np
        >>> from mindspore import dtype as mstype
        >>> from mindspore.nn.transformer import CrossEntropyLoss
        >>> from mindspore import Tensor
        >>> loss = CrossEntropyLoss()
        >>> logits = Tensor(np.array([[3, 5, 6, 9, 12, 33, 42, 12, 32, 72]]), mstype.float32)
        >>> labels_np = np.array([1]).astype(np.int32)
        >>> input_mask = Tensor(np.ones(1).astype(np.float32))
        >>> labels = Tensor(labels_np)
        >>> output = loss(logits, labels, input_mask)
        >>> print(output.shape)
        (1,)
    """

    def __init__(self, parallel_config=default_dpmp_config):
        super(CrossEntropyLoss, self).__init__()
        if _get_parallel_mode() in (ParallelMode.AUTO_PARALLEL,) and _is_sharding_propagation():
            if not isinstance(parallel_config, OpParallelConfig):
                raise TypeError("For 'CrossEntropyLoss', the class variable 'parallel_config' must be OpParallelConfig"
                                ", but got the type: {}.".format(type(parallel_config)))
            dp = parallel_config.data_parallel
            mp = parallel_config.model_parallel
            self.sum = P.ReduceSum()
            self.onehot = P.OneHot()
            # on/off value for onehot, for smooth labeling, modify the off_value
            self.on_value = Tensor(1.0, mstype.float32)
            self.off_value = Tensor(0.0, mstype.float32)
            self.max = P.ArgMaxWithValue(axis=-1, keep_dims=True).shard(
                ((dp, mp),))
            self.eps_const = Tensor(1e-24, mstype.float32)
            self.sub = P.Sub()
            self.exp = P.Exp()
            self.div = P.RealDiv()
            self.log = P.Log()
            self.add = P.Add()
            self.mul = P.Mul()
            self.neg = P.Neg()
            self.sum2 = P.ReduceSum().shard(((1,),))

            self.mul2 = P.Mul().shard(((1,), (1,)))
            self.add2 = P.Add()
            self.div2 = P.RealDiv()
        else:
            if not isinstance(parallel_config, OpParallelConfig):
                raise TypeError("For 'CrossEntropyLoss', the class variable 'parallel_config' must be OpParallelConfig"
                                ", but got the type: {}.".format(type(parallel_config)))
            dp = parallel_config.data_parallel
            mp = parallel_config.model_parallel
            self.sum = P.ReduceSum().shard(((dp, mp),))
            self.onehot = P.OneHot().shard(((dp, mp), (), ()))
            # on/off value for onehot, for smooth labeling, modify the off_value
            self.on_value = Tensor(1.0, mstype.float32)
            self.off_value = Tensor(0.0, mstype.float32)
            self.max = P.ArgMaxWithValue(axis=-1, keep_dims=True).shard(
                ((dp, mp),))
            self.eps_const = Tensor(1e-24, mstype.float32)
            self.sub = P.Sub().shard(((dp, mp), (dp, 1)))
            self.exp = P.Exp().shard(((dp, mp),))
            self.div = P.RealDiv().shard(((dp, mp), (dp, 1)))
            self.log = P.Log().shard(((dp, mp),))
            self.add = P.Add().shard(((dp, mp), ()))
            self.mul = P.Mul().shard(
                ((dp, mp), (dp, mp)))
            self.neg = P.Neg().shard(((dp, mp),))
            self.sum2 = P.ReduceSum().shard(((1,),))

            self.mul2 = P.Mul().shard(((1,), (1,)))
            self.add2 = P.Add()
            self.div2 = P.RealDiv()

    def construct(self, logits, label, input_mask):
        self._check_input(logits, label, input_mask)

        # the shape is [bs*seq_length, vocab_size]
        logits = F.cast(logits, mstype.float32)
        # LogSoftmax for logits over last dimension
        _, logit_max = self.max(logits)
        logit_sub = self.sub(logits, logit_max)
        logit_exp = self.exp(logit_sub)
        exp_sum = self.sum(logit_exp, -1)
        exp_sum = P.Reshape()(exp_sum, (F.shape(exp_sum)[0], 1))
        softmax_result = self.div(logit_exp, exp_sum)
        log_softmax_result = self.log(self.add(softmax_result, self.eps_const))

        # Flatten label to [bs*seq_length]
        label = P.Reshape()(label, (-1,))
        # Get onehot label [bs*seq_length, vocab_size]
        one_hot_label = self.onehot(label, F.shape(logits)[-1], self.on_value,
                                    self.off_value)
        # Cross-Entropy loss
        loss = self.mul(log_softmax_result, one_hot_label)
        loss_unsum = self.neg(loss)
        loss_reduce = self.sum(loss_unsum, -1)
        # input_mask indicates whether there is padded inputs and for padded inputs it will not be counted into loss
        input_mask = P.Reshape()(input_mask, (-1,))
        numerator = self.sum2(self.mul2(loss_reduce, input_mask))

        denominator = self.add2(
            self.sum2(input_mask),
            P.Cast()(F.tuple_to_array((1e-5,)), mstype.float32))
        loss = self.div2(numerator, denominator)
        return loss

    def _check_input(self, logits, label, input_mask):
        r"""Check the input tensor shape and type"""
        _check_is_tensor('logits', logits, self.cls_name)
        _check_is_tensor('label', label, self.cls_name)
        _check_is_tensor('input_mask', input_mask, self.cls_name)
        _check_input_dtype(F.dtype(logits), "logits", [mstype.float32, mstype.float16], self.cls_name)
        _check_input_dtype(F.dtype(label), "label", [mstype.int32], self.cls_name)
        _check_input_dtype(F.dtype(input_mask), "input_mask", [mstype.float32], self.cls_name)
        _check_input_shape(F.shape(logits), "logits", self.cls_name, 2)
        _check_input_shape(F.shape(label), "label", self.cls_name, 1)
        _check_input_shape(F.shape(input_mask), "input_mask", self.cls_name, 1)
        return True



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""
Functional Cells used in  finetune and evaluation.
"""

import os
import math
import numpy as np
import mindspore.nn as nn
from loguru import logger
from mindspore.ops import operations as P
from mindspore.common.tensor import Tensor
from mindspore.common import dtype as mstype
from mindspore.train.callback import Callback
from mindspore.nn.learning_rate_schedule import LearningRateSchedule, PolynomialDecayLR, WarmUpLR


class CrossEntropyCalculation(nn.Cell):
    """
    Cross Entropy loss
    """
    def __init__(self, is_training=True):
        super(CrossEntropyCalculation, self).__init__()
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, mstype.float32)
        self.off_value = Tensor(0.0, mstype.float32)
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.reshape = P.Reshape()
        self.last_idx = (-1,)
        self.neg = P.Neg()
        self.cast = P.Cast()
        self.is_training = is_training

    def construct(self, logits, label_ids, num_labels):
        if self.is_training:
            label_ids = self.reshape(label_ids, self.last_idx)
            one_hot_labels = self.onehot(label_ids, num_labels, self.on_value, self.off_value)
            per_example_loss = self.neg(self.reduce_sum(one_hot_labels * logits, self.last_idx))
            loss = self.reduce_mean(per_example_loss, self.last_idx)
            return_value = self.cast(loss, mstype.float32)
        else:
            return_value = logits * 1.0
        return return_value


def make_directory(path: str):
    """Make directory."""
    if path is None or not isinstance(path, str) or path.strip() == "":
        logger.error("The path(%r) is invalid type.", path)
        raise TypeError("Input path is invalid type")

    # convert the relative paths
    path = os.path.realpath(path)
    logger.debug("The abs path is %r", path)

    # check the path is exist and write permissions?
    if os.path.exists(path):
        real_path = path
    else:
        # All exceptions need to be caught because create directory maybe have some limit(permissions)
        logger.debug("The directory(%s) doesn't exist, will create it", path)
        try:
            os.makedirs(path, exist_ok=True)
            real_path = path
        except PermissionError as e:
            logger.error("No write permission on the directory(%r), error = %r", path, e)
            raise TypeError("No write permission on the directory.")
    return real_path

class LossCallBack(Callback):
    """
    Monitor the loss in training.
    If the loss in NAN or INF terminating training.
    Note:
        if per_print_times is 0 do not print loss.
    Args:
        per_print_times (int): Print loss every times. Default: 1.
    """
    def __init__(self, dataset_size=-1):
        super(LossCallBack, self).__init__()
        self._dataset_size = dataset_size
    def step_end(self, run_context):
        """
        Print loss after each step
        """
        cb_params = run_context.original_args()
        if self._dataset_size > 0:
            percent, epoch_num = math.modf(cb_params.cur_step_num / self._dataset_size)
            if percent == 0:
                percent = 1
                epoch_num -= 1
            print("epoch: {}, current epoch percent: {}, step: {}, outputs are {}"
                  .format(int(epoch_num), "%.3f" % percent, cb_params.cur_step_num, str(cb_params.net_outputs)),
                  flush=True)
        else:
            print("epoch: {}, step: {}, outputs are {}".format(cb_params.cur_epoch_num, cb_params.cur_step_num,
                                                               str(cb_params.net_outputs)), flush=True)

def LoadNewestCkpt(load_finetune_checkpoint_dir, steps_per_epoch, epoch_num, prefix):
    """
    Find the ckpt finetune generated and load it into eval network.
    """
    files = os.listdir(load_finetune_checkpoint_dir)
    pre_len = len(prefix)
    max_num = 0
    for filename in files:
        name_ext = os.path.splitext(filename)
        if name_ext[-1] != ".ckpt":
            continue
        if filename.find(prefix) == 0 and not filename[pre_len].isalpha():
            index = filename[pre_len:].find("-")
            if index == 0 and max_num == 0:
                load_finetune_checkpoint_path = os.path.join(load_finetune_checkpoint_dir, filename)
            elif index not in (0, -1):
                name_split = name_ext[-2].split('_')
                if (steps_per_epoch != int(name_split[len(name_split)-1])) \
                        or (epoch_num != int(filename[pre_len + index + 1:pre_len + index + 2])):
                    continue
                num = filename[pre_len + 1:pre_len + index]
                if int(num) > max_num:
                    max_num = int(num)
                    load_finetune_checkpoint_path = os.path.join(load_finetune_checkpoint_dir, filename)
    return load_finetune_checkpoint_path


class LearningRate(LearningRateSchedule):
    """
    Warmup-decay learning rate for  network.
    """
    def __init__(self, learning_rate, end_learning_rate, warmup_steps, decay_steps, power):
        super(LearningRate, self).__init__()
        self.warmup_flag = False
        if warmup_steps > 0:
            self.warmup_flag = True
            self.warmup_lr = WarmUpLR(learning_rate, warmup_steps)
        self.decay_lr = PolynomialDecayLR(learning_rate, end_learning_rate, decay_steps, power)
        self.warmup_steps = Tensor(np.array([warmup_steps]).astype(np.float32))

        self.greater = P.Greater()
        self.one = Tensor(np.array([1.0]).astype(np.float32))
        self.cast = P.Cast()

    def construct(self, global_step):
        decay_lr = self.decay_lr(global_step)
        if self.warmup_flag:
            is_warmup = self.cast(self.greater(self.warmup_steps, global_step), mstype.float32)
            warmup_lr = self.warmup_lr(global_step)
            lr = (self.one - is_warmup) * decay_lr + is_warmup * warmup_lr
        else:
            lr = decay_lr
        return lr


class LearningRateEpoch(LearningRateSchedule):
    """
    Warmup-decay learning rate for  network.
    """
    def __init__(self, steps_per_epoch, lr_steps, lr, weight_decay):
        super(LearningRate, self).__init__()
        self.steps_per_epoch = steps_per_epoch
        self.lr_steps = lr_steps
        self.lr = lr
        self.weight_decay = weight_decay

        self.greater = P.Greater()
        self.one = Tensor(np.array([1.0]).astype(np.float32))
        self.cast = P.Cast()

    def construct(self, global_step):
        epoch = global_step // self.steps_per_epoch
        decay = 0.1 ** (sum(epoch >= np.array(self.lr_steps)))
        lr = self.lr * decay
        decay = self.weight_decay
        return lr


def _get_poly_lr(global_step, lr_init, lr_end, lr_max, warmup_steps, total_steps, poly_power):
    """
    generate learning rate array
    Args:
       global_step(int): current step
       lr_init(float): init learning rate
       lr_end(float): end learning rate
       lr_max(float): max learning rate
       warmup_steps(int): number of warmup epochs
       total_steps(int): total epoch of training
       poly_power(int): poly learning rate power
    Returns:
       np.array, learning rate array
    """
    lr_each_step = []
    if warmup_steps != 0:
        inc_each_step = (float(lr_max) - float(lr_init)) / float(warmup_steps)
    else:
        inc_each_step = 0
    for i in range(total_steps):
        if i < warmup_steps:
            lr = float(lr_init) + inc_each_step * float(i)
        else:
            base = (1.0 - (float(i) - float(warmup_steps)) / (float(total_steps) - float(warmup_steps)))
            lr = float(lr_max - lr_end) * (base ** poly_power)
            lr = lr + lr_end
            if lr < 0.0:
                lr = 0.0
        lr_each_step.append(lr)

    learning_rate = np.array(lr_each_step).astype(np.float32)
    current_step = global_step
    learning_rate = learning_rate[current_step:]
    return learning_rate


def get__thor_lr(lr_max=0.0034, lr_min=3.244e-05, lr_power=1.0, lr_total_steps=30000):
    learning_rate = _get_poly_lr(global_step=0, lr_init=0.0, lr_end=lr_min, lr_max=lr_max, warmup_steps=0,
                                 total_steps=lr_total_steps, poly_power=lr_power)
    return Tensor(learning_rate)


def get__thor_damping(damping_max=5e-2, damping_min=1e-6, damping_power=1.0, damping_total_steps=30000):
    damping = _get_poly_lr(global_step=0, lr_init=0.0, lr_end=damping_min, lr_max=damping_max, warmup_steps=0,
                           total_steps=damping_total_steps, poly_power=damping_power)
    return Tensor(damping)

def get_file_list(input_file):
    """get file list of input folder"""
    if os.path.isdir(input_file):
        file_list = []
        for root, _, files in os.walk(input_file):
            for f in files:
                file_list.append(os.path.join(root, f))
        return file_list
    if os.path.isfile(input_file):
        return [input_file]
    raise ValueError('The input path: "{}" is not a folder or file.'.format(input_file))



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import numpy as np

import mindspore.nn as nn
from mindspore.common.initializer import initializer, TruncatedNormal
from mindspore.ops import operations as P
from mindspore.ops import functional as F
from mindspore.ops import composite as C
from mindspore.common.tensor import Tensor
from mindspore.common.parameter import Parameter
from mindspore.common.api import ms_function
from mindspore.common import dtype as mstype
from mindspore.nn.wrap.grad_reducer import DistributedGradReducer
from mindspore.context import ParallelMode
from mindspore.communication.management import get_group_size
from mindspore.parallel._utils import (_get_device_num, _get_gradients_mean,
                                       _get_parallel_mode, _get_enable_parallel_optimizer)
from mindspore import context
import mindspore as ms

# from wywLM.optims.utils import LossCallBack
from .models import BertModel
from . import msops

GRADIENT_CLIP_TYPE = 1
GRADIENT_CLIP_VALUE = 1.0

clip_grad = C.MultitypeFuncGraph("clip_grad")


@clip_grad.register("Number", "Number", "Tensor")
def _clip_grad(clip_type, clip_value, grad):
    """
    Clip gradients.

    Inputs:
        clip_type (int): The way to clip, 0 for 'value', 1 for 'norm'.
        clip_value (float): Specifies how much to clip.
        grad (tuple[Tensor]): Gradients.

    Outputs:
        tuple[Tensor], clipped gradients.
    """
    if clip_type not in (0, 1):
        return grad
    dt = F.dtype(grad)
    if clip_type == 0:
        new_grad = C.clip_by_value(grad, F.cast(F.tuple_to_array((-clip_value,)), dt),
                                   F.cast(F.tuple_to_array((clip_value,)), dt))
    else:
        new_grad = nn.ClipByNorm()(grad, F.cast(F.tuple_to_array((clip_value,)), dt))
    return new_grad


class NetworkWithLoss(nn.Cell):
    """
    Provide loss through network.

    Args:
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings. Default: False.

    Returns:
        Tensor, the loss of the network.
    """

    def __init__(self, backbone, loss_fn, return_all=False, fp16=False):
        super(NetworkWithLoss, self).__init__()
        self.backbone = backbone
        self.loss_fn = loss_fn
        if loss_fn.construct.__code__.co_argcount == 3:
            self.use_simple_loss = True
        else:
            self.use_simple_loss = False
        self.fp16 = fp16
        self.return_all = return_all

    def construct(self, *sample):
        """Get pre-training loss"""
        prediction_scores = self.backbone(*sample)
        if self.use_simple_loss:
            total_loss = self.loss_fn(*(prediction_scores + (sample[1],)))
        else:
            total_loss = self.loss_fn(*(sample + prediction_scores))
        total_loss = F.cast(total_loss, mstype.float32)
        return total_loss


class ModelForEval(nn.Cell):
    '''
    Evaluate scores
    '''
    def __init__(self, network, eval_head=None, fp16=False):
        super(ModelForEval, self).__init__(auto_prefix=False)
        self.eval_head = eval_head
        self.backbone = network.backbone
        if eval_head is None:
            self.eval_head = network.loss_fn
        if eval_head.construct.__code__.co_argcount == 3:
            self.use_simple_eval = True
        else:
            self.use_simple_eval = False
        self.fp16 = fp16
        self.cast = ms.ops.Cast()
        self.reduce_flag = False
        parallel_mode = context.get_auto_parallel_context("parallel_mode")
        if parallel_mode in [ParallelMode.DATA_PARALLEL, ParallelMode.HYBRID_PARALLEL] and context.get_context("mode") == context.GRAPH_MODE:
            self.reduce_flag = True
            # self.allreduce = P.AllReduce()
            self.allgather = AllGather() #ms.ops.AllGather()
        else:
            self.allgather = F.identity

    def construct(self, *sample):
        """Calculate prediction scores"""
        pred_scores = self.backbone(*sample)
        if self.use_simple_eval:
            output = self.eval_head(*(pred_scores + (sample[1],)))
        else:
            output = self.eval_head(*(sample + pred_scores))
        out = ()
        for m in output:
            if m.ndim == 0:
                m = F.expand_dims(m, 0)
                out += (self.allgather(m),)
            else:
                out += (self.allgather(m),)
        return out


class TrainOneStepCell(nn.TrainOneStepCell):
    """
    Encapsulation class of network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        sens (Number): The adjust parameter. Default: 1.0.
        enable_clip_grad (boolean): If True, clip gradients in BertTrainOneStepCell. Default: True.
    """

    def __init__(self, network, optimizer, sens=1.0, enable_clip_grad=False):
        super(TrainOneStepCell, self).__init__(network, optimizer, sens)
        # self.cast = P.Cast()
        self.hyper_map = C.HyperMap()
        self.enable_clip_grad = enable_clip_grad
        self.enable_tuple_broaden = True

    def set_sens(self, value):
        self.sens = value

    @ms_function
    def clip_grads(self, grads):
        grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        return grads

    def construct(self, *args):
        """Defines the computation performed."""
        weights = self.weights

        loss = self.network(*args)
        grads = self.grad(self.network, weights)(*args,
                                                 F.cast(F.tuple_to_array((self.sens,)),
                                                           mstype.float32))
        if self.enable_clip_grad:
            grads = self.clip_grads(grads)
        grads = self.grad_reducer(grads)
        self.optimizer(grads)
        return loss


grad_scale = C.MultitypeFuncGraph("grad_scale")
reciprocal = P.Reciprocal()


@grad_scale.register("Tensor", "Tensor")
def tensor_grad_scale(scale, grad):
    return grad * reciprocal(scale)


_grad_overflow = C.MultitypeFuncGraph("_grad_overflow")
grad_overflow = P.FloatStatus()


@_grad_overflow.register("Tensor")
def _tensor_grad_overflow(grad):
    return grad_overflow(grad)


class TrainOneStepWithLossScaleCell(nn.TrainOneStepWithLossScaleCell):
    """
    Encapsulation class of network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
    """

    def __init__(self, network, optimizer, scale_update_cell=None, enable_clip_grad=True, opt_overflow=False):
        super(TrainOneStepWithLossScaleCell, self).__init__(network, optimizer, scale_update_cell)
        # self.cast = P.Cast()
        self.degree = 1
        # if self.reducer_flag:
        #     self.degree = get_group_size()
        #     self.grad_reducer = DistributedGradReducer(optimizer.parameters, False, self.degree)
        self.enable_clip_grad = enable_clip_grad
        
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=mstype.float32))
        else:
            self.loss_scale = Parameter(Tensor(np.array([1]).astype(np.int32)), requires_grad=False)
        self.enable_tuple_broaden = True
        self.opt_overflow = opt_overflow
        self.one = Tensor(np.array([1]).astype(np.int32))
        self.zero = Tensor(np.array([0]).astype(np.int32))

    @ms_function
    def clip_grads(self, grads):
        if self.enable_clip_grad:
            grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        else:
            grads = F.identity(grads)
        return grads

    def construct(self, *args):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(*args)

        # if sens is None:
        #     scaling_sens = self.loss_scale
        # else:
        #     scaling_sens = self.one
        scaling_sens = self.loss_scale
        
        status, scaling_sens = self.start_overflow_check(loss, scaling_sens)
        grad_fn = self.grad(self.network, weights)
        grads = grad_fn(*args, F.cast(scaling_sens, mstype.float32))

        # apply grad reducer on grads
        grads = self.grad_reducer(grads)

        degree_sens = F.cast(scaling_sens * self.degree, mstype.float32)
        grads = self.hyper_map(F.partial(grad_scale, degree_sens), grads)
        
        grads = self.clip_grads(grads)

        cond = self.get_overflow_status(status, grads)
        
        # if sens is None:
        #     overflow = self.loss_scaling_manager(self.loss_scale, cond)
        # else:
        #     overflow = cond
        overflow = self.loss_scaling_manager(self.loss_scale, cond)
        # if self.opt_overflow:
        if not overflow:
            self.optimizer(grads)
        # else:
        #     # if not overflow.asnumpy().tolist():
        #     self.optimizer(grads)

        return (loss, cond, scaling_sens)


class BertTrainOneStepWithLossScaleCell(nn.TrainOneStepWithLossScaleCell):
    """
    Encapsulation class of bert network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
    """

    def __init__(self, network, optimizer, scale_update_cell=None):
        super(BertTrainOneStepWithLossScaleCell, self).__init__(network, optimizer, scale_update_cell)
        self.cast = P.Cast()
        self.degree = 1
        if self.reducer_flag:
            self.degree = get_group_size()
            self.grad_reducer = DistributedGradReducer(optimizer.parameters, False, self.degree)

        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=mstype.float32))
        self.enable_tuple_broaden = True

    @ms_function
    def clip_grads(self, grads):
        grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        return grads

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            next_sentence_labels,
                            masked_lm_positions,
                            masked_lm_ids,
                            masked_lm_weights)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens
        status, scaling_sens = self.start_overflow_check(loss, scaling_sens)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 next_sentence_labels,
                                                 masked_lm_positions,
                                                 masked_lm_ids,
                                                 masked_lm_weights,
                                                 self.cast(scaling_sens,
                                                           mstype.float32))
        # apply grad reducer on grads
        grads = self.grad_reducer(grads)
        degree_sens = self.cast(scaling_sens * self.degree, mstype.float32)
        grads = self.hyper_map(F.partial(grad_scale, degree_sens), grads)
        grads = self.clip_grads(grads)

        cond = self.get_overflow_status(status, grads)
        overflow = cond
        if sens is None:
            overflow = self.loss_scaling_manager(self.loss_scale, cond)
        if not overflow:
            self.optimizer(grads)
        return (loss, cond, scaling_sens)


class TrainOneStepWithLossScaleCellForAdam(nn.TrainOneStepWithLossScaleCell):
    """
    Encapsulation class of network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.
    Different from TrainOneStepWithLossScaleCell, the optimizer takes the overflow
    condition as input.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
    """
    def __init__(self, network, optimizer, scale_update_cell=None, enable_clip_grad=True):
        super(TrainOneStepWithLossScaleCellForAdam, self).__init__(network, optimizer, scale_update_cell)
        self.cast = P.Cast()
        # self.degree = 1
        # if self.reducer_flag:
        #     self.degree = get_group_size()
        #     self.grad_reducer = DistributedGradReducer(optimizer.parameters, False, self.degree)
        self.enable_clip_grad = enable_clip_grad
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=mstype.float32))
        self.enable_tuple_broaden = True

    @ms_function
    def clip_grads(self, grads):
        grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        return grads

    def construct(self, *args, sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(*args)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens

        status, scaling_sens = self.start_overflow_check(loss, scaling_sens)
        grads = self.grad(self.network, weights)(*args,
                                                 self.cast(scaling_sens,
                                                           mstype.float32))
        # apply grad reducer on grads
        grads = self.grad_reducer(grads)
        grads = self.hyper_map(F.partial(grad_scale, scaling_sens * self.degree), grads)
        if self.enable_clip_grad:
            grads = self.clip_grads(grads)
        cond = self.get_overflow_status(status, grads)
        overflow = cond
        if self.loss_scaling_manager is not None:
            overflow = self.loss_scaling_manager(scaling_sens, cond)
        self.optimizer(grads, overflow)
        return (loss, cond, scaling_sens)

cast = P.Cast()
add_grads = C.MultitypeFuncGraph("add_grads")


@add_grads.register("Tensor", "Tensor")
def _add_grads(accu_grad, grad):
    return accu_grad + cast(grad, mstype.float32)

update_accu_grads = C.MultitypeFuncGraph("update_accu_grads")

@update_accu_grads.register("Tensor", "Tensor")
def _update_accu_grads(accu_grad, grad):
    succ = True
    return F.depend(succ, F.assign(accu_grad, cast(grad, mstype.float32)))

accumulate_accu_grads = C.MultitypeFuncGraph("accumulate_accu_grads")

@accumulate_accu_grads.register("Tensor", "Tensor")
def _accumulate_accu_grads(accu_grad, grad):
    succ = True
    return F.depend(succ, F.assign_add(accu_grad, cast(grad, mstype.float32)))


zeroslike = P.ZerosLike()
reset_accu_grads = C.MultitypeFuncGraph("reset_accu_grads")


@reset_accu_grads.register("Tensor")
def _reset_accu_grads(accu_grad):
    succ = True
    return F.depend(succ, F.assign(accu_grad, zeroslike(accu_grad)))


class TrainAccumulationAllReducePostWithLossScaleCell(nn.Cell):
    """
    Encapsulation class of network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    To mimic higher batch size, gradients are accumulated N times before weight update.

    For distribution mode, allreduce will only be implemented in the weight updated step,
    i.e. the sub-step after gradients accumulated N times.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
        accumulation_steps (int): Number of accumulation steps before gradient update. The global batch size =
                                batch_size * accumulation_steps. Default: 1.
    """

    def __init__(self, network, optimizer, scale_update_cell=None, accumulation_steps=1, 
                    enable_global_norm=False, opt_overflow=False, gpu_target=False):
        super(TrainAccumulationAllReducePostWithLossScaleCell, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.accumulation_steps = accumulation_steps
        self.enable_global_norm = enable_global_norm
        self.one = Tensor(np.array([1]).astype(np.int32))
        self.zero = Tensor(np.array([0]).astype(np.int32))
        self.local_step = Parameter(initializer(0, [1], mstype.int32))
        self.accu_grads = self.weights.clone(prefix="accu_grads", init='zeros')
        self.accu_overflow = Parameter(initializer(0, [1], mstype.int32))
        self.accu_loss = Parameter(initializer(0, [1], mstype.float32))

        self.grad = C.GradOperation(get_by_list=True, sens_param=True)
        self.reducer_flag = False
        self.parallel_mode = context.get_auto_parallel_context("parallel_mode")
        if self.parallel_mode in [ParallelMode.DATA_PARALLEL, ParallelMode.HYBRID_PARALLEL]:
            self.reducer_flag = True
        self.grad_reducer = F.identity
        self.degree = 1
        if self.reducer_flag:
            self.degree = _get_device_num()
            self.mean = _get_gradients_mean()
            self.grad_reducer = DistributedGradReducer(optimizer.parameters, self.mean, self.degree)
        self.is_distributed = (self.parallel_mode != ParallelMode.STAND_ALONE)
        self.overflow_reducer = F.identity
        if self.is_distributed:
            self.overflow_reducer = P.AllReduce()
        self.cast = P.Cast()
        self.alloc_status = P.NPUAllocFloatStatus()
        self.get_status = P.NPUGetFloatStatus()
        self.clear_status = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.base = Tensor(1, mstype.float32)
        self.less_equal = P.LessEqual()
        self.logical_or = P.LogicalOr()
        self.not_equal = P.NotEqual()
        self.select = P.Select()
        self.reshape = P.Reshape()
        self.hyper_map = C.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        self.use_loss_scale = Tensor(self.loss_scaling_manager is not None)
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=mstype.float32))

        self.opt_overflow = opt_overflow
        self.gpu_target = gpu_target

    def construct(self, *args): #, sens=1
        """Defines the computation performed."""
        # start = time.time()
        weights = self.weights
        
        loss = self.network(*args)
        
        # self.last_weights = [w.asnumpy() for w in weights]
        # if self.not_equal(sens, 1):
        #     scaling_sens = sens
        # else:
        # loss = loss[0]
        if self.use_loss_scale:
            scaling_sens = self.loss_scale
            scaling_sens = C.ones_like(loss) * self.cast(scaling_sens, F.dtype(loss))
            scaling_sens = self.cast(scaling_sens, mstype.float32)
        else:
            scaling_sens = self.one

        # alloc status and clear should be right before gradoperation
        if not self.gpu_target:
            init = self.alloc_status()
            init = F.depend(init, loss)
            clear_status = self.clear_status(init)
            # if self.use_loss_scale:
            scaling_sens = F.depend(scaling_sens, clear_status)
        else:
            init = (False)

        # update accumulation parameters
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)
        self.local_step = self.select(is_accu_step, self.local_step + self.one, self.one)
        self.accu_loss = self.select(is_accu_step, self.accu_loss + loss, loss.view((1,)))
        mean_loss = self.accu_loss / self.local_step
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)

        grads_fn = self.grad(self.network, weights)
        grads = grads_fn(*args, scaling_sens)

        accu_succ = self.hyper_map(accumulate_accu_grads, self.accu_grads, grads)
        mean_loss = F.depend(mean_loss, accu_succ)
        if not self.gpu_target:
            init = F.depend(init, mean_loss)
            get_status = self.get_status(init)
            init = F.depend(init, get_status)
            flag_sum = self.reduce_sum(init, (0,))
            overflow = self.less_equal(self.base, flag_sum)
        else:
            flag_sum = self.hyper_map(self.partial(_grad_overflow), mean_loss)
            flag_sum = P.AddN()(flag_sum)
            # convert flag_sum to scalar
            flag_sum = P.Reshape()(flag_sum, (()))
            if self.is_distributed:
                flag_reduce = self.allreduce(flag_sum)
                overflow = self.less_equal(self.base, flag_reduce)
            else:
                overflow = self.less_equal(self.base, flag_sum)
        
        overflow = self.logical_or(self.not_equal(self.accu_overflow, self.zero), overflow)
        accu_overflow = self.select(overflow, self.one, self.zero)
        self.accu_overflow = self.select(is_accu_step, accu_overflow, self.zero)

        if is_accu_step:
            # apply grad reducer on grads
            grads = self.grad_reducer(self.accu_grads)
            scaling = scaling_sens * self.degree * self.accumulation_steps
            grads = self.hyper_map(F.partial(grad_scale, scaling), grads)
            # if self.enable_global_norm:
            #     grads = C.clip_by_global_norm(grads, 1.0, None)
            # else:
            #     grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
            accu_overflow = F.depend(accu_overflow, grads)
            accu_overflow = self.overflow_reducer(accu_overflow)
            overflow = self.less_equal(self.base, accu_overflow)
            accu_succ = self.hyper_map(reset_accu_grads, self.accu_grads)
            overflow = F.depend(overflow, accu_succ)
            overflow = self.reshape(overflow, (()))
            # if sens is None:
            if self.use_loss_scale:
                overflow = self.loss_scaling_manager(self.loss_scale, overflow)
                # else:
                #     overflow = False
            if self.opt_overflow:
                # self.last_grads = [g.asnumpy() for g in grads]
                # if np.any(np.isnan(self.last_grads[0])):
                #     print('error')
                self.optimizer(grads, overflow)
            else:
                if not (overflow and self.use_loss_scale):
                    self.optimizer(grads)
        # print(time.time() - start)
        return (mean_loss, overflow, scaling_sens)


class TrainAccumulationAllReduceEachWithLossScaleCell(nn.Cell):
    """
    Encapsulation class of network training.

    Append an optimizer to the training network after that the construct
    function can be called to create the backward graph.

    To mimic higher batch size, gradients are accumulated N times before weight update.

    For distribution mode, allreduce will be implemented after each sub-step and the trailing time
    will be overided by backend optimization pass.

    Args:
        network (Cell): The training network. Note that loss function should have been added.
        optimizer (Optimizer): Optimizer for updating the weights.
        scale_update_cell (Cell): Cell to do the loss scale. Default: None.
        accumulation_steps (int): Number of accumulation steps before gradient update. The global batch size =
                                  batch_size * accumulation_steps. Default: 1.
    """
    def __init__(self, network, optimizer, scale_update_cell=None, accumulation_steps=1, enable_global_norm=False):
        super(TrainAccumulationAllReduceEachWithLossScaleCell, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.accumulation_steps = accumulation_steps
        self.enable_global_norm = enable_global_norm
        self.one = Tensor(np.array([1]).astype(np.int32))
        self.zero = Tensor(np.array([0]).astype(np.int32))
        self.local_step = Parameter(initializer(0, [1], mstype.int32))
        self.accu_grads = self.weights.clone(prefix="accu_grads", init='zeros')
        self.accu_overflow = Parameter(initializer(0, [1], mstype.int32))
        self.accu_loss = Parameter(initializer(0, [1], mstype.float32))

        self.grad = C.GradOperation(get_by_list=True, sens_param=True)
        self.reducer_flag = False
        self.parallel_mode = context.get_auto_parallel_context("parallel_mode")
        if self.parallel_mode in [ParallelMode.DATA_PARALLEL, ParallelMode.HYBRID_PARALLEL]:
            self.reducer_flag = True
        self.grad_reducer = F.identity
        self.degree = 1
        if self.reducer_flag:
            self.degree = get_group_size()
            self.grad_reducer = DistributedGradReducer(optimizer.parameters, False, self.degree)
        self.is_distributed = (self.parallel_mode != ParallelMode.STAND_ALONE)
        self.overflow_reducer = F.identity
        if self.is_distributed:
            self.overflow_reducer = P.AllReduce()
        self.cast = P.Cast()
        self.alloc_status = P.NPUAllocFloatStatus()
        self.get_status = P.NPUGetFloatStatus()
        self.clear_before_grad = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.base = Tensor(1, mstype.float32)
        self.less_equal = P.LessEqual()
        self.logical_or = P.LogicalOr()
        self.not_equal = P.NotEqual()
        self.select = P.Select()
        self.reshape = P.Reshape()
        self.hyper_map = C.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_cell
        if scale_update_cell:
            self.loss_scale = Parameter(Tensor(scale_update_cell.get_loss_scale(), dtype=mstype.float32))

    @C.add_flags(has_effect=True)
    def construct(self,
                  *args,
                  sens=None):
        """Defines the computation performed."""
        weights = self.weights
        loss = self.network(*args)[0]
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens

        # update accumulation parameters
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)
        self.local_step = self.select(is_accu_step, self.local_step + self.one, self.one)
        self.accu_loss = self.select(is_accu_step, self.accu_loss + loss, loss)
        mean_loss = self.accu_loss / self.local_step
        is_accu_step = self.not_equal(self.local_step, self.accumulation_steps)

        # alloc status and clear should be right before gradoperation
        init = self.alloc_status()
        self.clear_before_grad(init)
        grads = self.grad(self.network, weights)(*args,
                                                 self.cast(scaling_sens,
                                                           mstype.float32))


        accu_grads = self.hyper_map(add_grads, self.accu_grads, grads)
        scaling = scaling_sens * self.degree * self.accumulation_steps
        grads = self.hyper_map(F.partial(grad_scale, scaling), accu_grads)
        grads = self.grad_reducer(grads)

        self.get_status(init)
        flag_sum = self.reduce_sum(init, (0,))
        flag_reduce = self.overflow_reducer(flag_sum)
        overflow = self.less_equal(self.base, flag_reduce)
        overflow = self.logical_or(self.not_equal(self.accu_overflow, self.zero), overflow)
        accu_overflow = self.select(overflow, self.one, self.zero)
        self.accu_overflow = self.select(is_accu_step, accu_overflow, self.zero)
        overflow = self.reshape(overflow, (()))

        if is_accu_step:
            succ = False
            accu_succ = self.hyper_map(update_accu_grads, self.accu_grads, accu_grads)
            succ = F.depend(succ, accu_succ)
        else:
            if sens is None:
                overflow = self.loss_scaling_manager(self.loss_scale, overflow)
            if overflow:
                succ = False
            else:
                if self.enable_global_norm:
                    grads = C.clip_by_global_norm(grads, 1.0, None)
                else:
                    grads = self.hyper_map(F.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)

                succ = self.optimizer(grads)

            accu_succ = self.hyper_map(reset_accu_grads, self.accu_grads)
            succ = F.depend(succ, accu_succ)

        ret = (mean_loss, overflow, scaling_sens)
        return F.depend(ret, succ)


class NetworkMatchBucket(nn.Cell):
    '''
    execute according to different sentence lengths.
    '''
    def __init__(self, network, seq_length, bucket_list=None):
        super(NetworkMatchBucket, self).__init__()
        self.network = network
        if not bucket_list or not isinstance(bucket_list, list):
            bucket_list = [seq_length]
        self.bucket_list = [bucket for bucket in bucket_list if bucket <= seq_length]

        if network.reducer_flag:
            reuse_attr = 'reuse_communication_node'
            if not network.grad_reducer.split_fusion:
                hccl_op = network.grad_reducer.allreduce
                network.grad_reducer.allreduce = hccl_op.add_prim_attr(reuse_attr, getattr(hccl_op, 'fusion'))
            else:
                new_op_list = []
                for hccl_op in network.grad_reducer.op_list:
                    new_op = hccl_op.add_prim_attr(reuse_attr, getattr(hccl_op, 'fusion'))
                    new_op_list.append(new_op)
                network.grad_reducer.op_list = new_op_list

    def construct(self,*args,
                  sentence_flag):
        """Switch network according to sentence length."""
        for bucket in self.bucket_list:
            if sentence_flag == bucket:
                input_ids = input_ids[:, :bucket]
                input_mask = input_mask[:, :bucket]
                token_type_id = token_type_id[:, :bucket]
                loss = self.network(*args)
                return loss

        loss = self.network(*args)
        return loss


class AllGather(nn.Cell):
    def __init__(self):
        super(AllGather, self).__init__()
        self.allgather = ms.ops.AllGather()
    def construct(self, x):
        return self.allgather(x)


def model_adapter(model, optimizer, scale_update_cell=None, accumulation_steps=1, 
                    enable_global_norm=False, opt_overflow=False, gpu_target=False):
    if accumulation_steps <= 1:
        if scale_update_cell is None:
            network = TrainOneStepCell(model, optimizer).set_train()
        else:
            network = TrainOneStepWithLossScaleCell(model, optimizer=optimizer, 
                                                    scale_update_cell=scale_update_cell, 
                                                    opt_overflow=opt_overflow).set_train()
    else:
        network = TrainAccumulationAllReducePostWithLossScaleCell(model, optimizer=optimizer,
                                                scale_update_cell=scale_update_cell,
                                                accumulation_steps=accumulation_steps,
                                                enable_global_norm=enable_global_norm,
                                                opt_overflow=opt_overflow,
                                                gpu_target=gpu_target).set_train()
    return network
    


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import math
import copy
from collections import Sequence
import mindspore

import numpy
import mindspore.numpy as np
import mindspore.common.dtype as mstype
import mindspore.nn as nn
import mindspore.ops.functional as F
from mindspore.common.initializer import TruncatedNormal, initializer
from mindspore.ops import operations as P
from mindspore.ops import composite as C
import mindspore.ops as O
from mindspore.common.tensor import Tensor
from mindspore.common.parameter import Parameter

class BertForClassify(nn.Cell):
    """
    Provide loss through network.

    Args:
        is_training (bool): Specifies whether to use the training mode.
        use_one_hot_embeddings (bool): Specifies whether to use one-hot for embeddings. Default: False.

    Returns:
        Tensor, the loss of the network.
    """

    def __init__(self, backbone, num_labels, return_all=False, fp16=False, dropout=0.1):
        super(BertForClassify, self).__init__()
        self.bert = backbone
        self.classifier = nn.Dense(self.bert.hidden_size, num_labels, has_bias=False)
        # self.loss_fn = nn.SoftmaxCrossEntropyWithLogits()
        self.dropout = nn.Dropout(1 - dropout)
        self.fp16 = fp16
        self.return_all = return_all

    def construct(self, *sample):
        input_ids, input_mask, token_type_id = sample[: 3]
        
        prediction_scores = self.bert(input_ids, token_type_id, input_mask)
        pooled_out = F.cast(prediction_scores[1], mstype.float32)
        logits = self.classifier(self.dropout(pooled_out))
        # total_loss = self.loss_fn(logits, label)
        # total_loss = F.cast(total_loss, mstype.float32)
        return (logits,)


class ClsLoss(nn.SoftmaxCrossEntropyWithLogits):
    def __init__(self, sparse=True, reduction='none'):
        super().__init__(sparse=sparse, reduction=reduction)
    
    def construct(self, *inputs):
        logits = inputs[-1]
        labels = inputs[-2]
        return super().construct(logits, labels)


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from .trainer import DistributedTrainer, set_random_seed
from .args import get_args
# from .dist_launcher import initialize_distributed,kill_children
# from ._utils import batch_to,batch_apply



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from distutils.command.config import config
import math
import copy
from collections import Sequence
import mindspore

import numpy
import mindspore.numpy as np
import mindspore.common.dtype as mstype
import mindspore.nn as nn
import mindspore.ops.functional as F
from mindspore.common.initializer import TruncatedNormal, initializer
from mindspore.ops import operations as P
from mindspore.ops import composite as C
import mindspore.ops as O
from mindspore.common.tensor import Tensor
from mindspore.common.parameter import Parameter
from mindspore import ms_function

from MindsporeTrainer.modeling.config import ModelConfig


class CreateAttentionMaskFromInputMask(nn.Cell):
    """
    Create attention mask according to input mask.

    Args:
        config (Class): Configuration for BertModel.
    """
    def __init__(self):
        super(CreateAttentionMaskFromInputMask, self).__init__()
        self.input_mask = None

        self.cast = P.Cast()
        self.reshape = P.Reshape()

    def construct(self, input_mask):
        seq_length = F.shape(input_mask)[1]
        attention_mask = self.cast(self.reshape(input_mask, (-1, 1, seq_length)), mstype.float32)
        return attention_mask


class CellList(nn.CellList):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def construct(self, *inputs):
        for cell in self._cells.values():
            inputs = (cell(*inputs),)
        return inputs[0]


class Embeddings(nn.Cell):
    def __init__(self,
                 use_one_hot_embeddings=False,
                 **kwargs):
        super().__init__()
        vocab_size = kwargs.get('vocab_size')
        embedding_size = kwargs.get('hidden_size')
        hidden_size = kwargs.get('hidden_size')
        type_vocab_size = kwargs.get('type_vocab_size')
        hidden_dropout_prob = kwargs.get('hidden_dropout_prob')
        padding_idx = kwargs.get('padding_idx')
        max_position_embeddings = kwargs.get('max_position_embeddings')
        self.embedding_size = embedding_size
        self.hidden_size = hidden_size
        self.use_relative_positions = kwargs.get('use_relative_positions')
        self.word_embeddings = nn.Embedding(vocab_size, embedding_size, padding_idx=padding_idx, use_one_hot=use_one_hot_embeddings,
                                            embedding_table='truncatedNormal')
        self.position_biased_input = kwargs.get('position_biased_input')
        if type_vocab_size > 0:
            self.use_token_type = True
            self.token_type_embeddings = nn.Embedding(type_vocab_size, embedding_size)
        else:
            self.use_token_type = False
        self.type_vocab_size = type_vocab_size
        if embedding_size != hidden_size:
            self.embed_proj = nn.Dense(embedding_size, hidden_size, has_bias=False)
        self.position_embeddings = nn.Embedding(
            vocab_size=max_position_embeddings,
            embedding_size=embedding_size,
            use_one_hot=False)
        ln_eps = kwargs.get('layer_norm_eps', 1e-7)
        self.LayerNorm = nn.LayerNorm((hidden_size, ), epsilon=ln_eps)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.add = P.Add()

    def construct(self, input_ids, token_type_ids, mask=None, **kwargs):
        seq_length = input_ids.shape[1]

        position_ids = np.arange(0, seq_length).astype(mstype.int32)
        # position_ids = Tensor(np.arange(0, seq_length), dtype=mstype.int32)
        position_ids = O.BroadcastTo(input_ids.shape)(F.expand_dims(position_ids, 0))

        # if token_type_ids is None:
        #     token_type_ids = O.ZerosLike()(input_ids)

        output = self.word_embeddings(input_ids)
        
        if self.use_token_type:
            token_type_embeddings = self.token_type_embeddings(token_type_ids)
            output = self.add(output, token_type_embeddings)
        # else:
        #     output = self.add(output, 0)
        
        # if not self.use_relative_positions:
        position_embeddings = self.position_embeddings(position_ids)
        output = self.add(output, position_embeddings)
        # else:
        #     position_embeddings = None

        if self.embedding_size != self.hidden_size:
            output = self.embed_proj(output)
        # else:
        #     output = self.add(output, 0)
        embeddings = self.LayerNorm(output)
        embeddings = self.dropout(embeddings)
        return {
            'embeddings': embeddings,
            'position_embeddings': position_embeddings}   

class EmbeddingLookup(nn.Cell):
    """
    A embeddings lookup table with a fixed dictionary and size.

    Args:
        vocab_size (int): Size of the dictionary of embeddings.
        embedding_size (int): The size of each embedding vector.
        embedding_shape (list): [batch_size, seq_length, embedding_size], the shape of
                         each embedding vector.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
    """
    def __init__(self,
                 vocab_size,
                 embedding_size,
                 embedding_shape,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02):
        super(EmbeddingLookup, self).__init__()
        self.vocab_size = vocab_size
        self.use_one_hot_embeddings = use_one_hot_embeddings
        self.embedding_table = Parameter(initializer
                                         (TruncatedNormal(initializer_range),
                                          [vocab_size, embedding_size]))
        self.expand = P.ExpandDims()
        self.shape_flat = (-1,)
        self.gather = P.Gather()
        self.one_hot = P.OneHot()
        self.on_value = Tensor(1.0, mstype.float32)
        self.off_value = Tensor(0.0, mstype.float32)
        self.array_mul = P.MatMul()
        self.reshape = P.Reshape()
        self.shape = tuple(embedding_shape)

    def construct(self, input_ids):
        """Get output and embeddings lookup table"""
        extended_ids = self.expand(input_ids, -1)
        flat_ids = self.reshape(extended_ids, self.shape_flat)
        if self.use_one_hot_embeddings:
            one_hot_ids = self.one_hot(flat_ids, self.vocab_size, self.on_value, self.off_value)
            output_for_reshape = self.array_mul(
                one_hot_ids, self.embedding_table)
        else:
            output_for_reshape = self.gather(self.embedding_table, flat_ids, 0)
        output = self.reshape(output_for_reshape, self.shape)
        return output, self.embedding_table


class EmbeddingPostprocessor(nn.Cell):
    """
    Postprocessors apply positional and token type embeddings to word embeddings.

    Args:
        embedding_size (int): The size of each embedding vector.
        embedding_shape (list): [batch_size, seq_length, embedding_size], the shape of
                         each embedding vector.
        use_token_type (bool): Specifies whether to use token type embeddings. Default: False.
        token_type_vocab_size (int): Size of token type vocab. Default: 16.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        max_position_embeddings (int): Maximum length of sequences used in this
                                 model. Default: 512.
        dropout_prob (float): The dropout probability. Default: 0.1.
    """
    def __init__(self,
                 embedding_size,
                 embedding_shape,
                 use_relative_positions=False,
                 use_token_type=False,
                 token_type_vocab_size=16,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 max_position_embeddings=512,
                 dropout_prob=0.1):
        super(EmbeddingPostprocessor, self).__init__()
        self.use_token_type = use_token_type
        self.token_type_vocab_size = token_type_vocab_size
        self.use_one_hot_embeddings = use_one_hot_embeddings
        self.max_position_embeddings = max_position_embeddings
        self.token_type_embedding = nn.Embedding(
            vocab_size=token_type_vocab_size,
            embedding_size=embedding_size,
            use_one_hot=use_one_hot_embeddings)
        self.shape_flat = (-1,)
        self.one_hot = P.OneHot()
        self.on_value = Tensor(1.0, mstype.float32)
        self.off_value = Tensor(0.1, mstype.float32)
        self.array_mul = P.MatMul()
        self.reshape = P.Reshape()
        self.shape = tuple(embedding_shape)
        self.dropout = nn.Dropout(1 - dropout_prob)
        self.gather = P.Gather()
        self.use_relative_positions = use_relative_positions
        self.slice = P.StridedSlice()
        _, seq, _ = self.shape
        self.full_position_embedding = nn.Embedding(
            vocab_size=max_position_embeddings,
            embedding_size=embedding_size,
            use_one_hot=False)
        self.layernorm = nn.LayerNorm((embedding_size,))
        self.position_ids = Tensor(np.arange(seq).reshape(-1, seq).astype(np.int32))
        self.add = P.Add()

    def construct(self, token_type_ids, word_embeddings):
        """Postprocessors apply positional and token type embeddings to word embeddings."""
        output = word_embeddings
        if self.use_token_type:
            token_type_embeddings = self.token_type_embedding(token_type_ids)
            output = self.add(output, token_type_embeddings)
        if not self.use_relative_positions:
            shape = F.shape(output)
            position_ids = self.position_ids[:, :shape[1]]
            position_embeddings = self.full_position_embedding(position_ids)
            output = self.add(output, position_embeddings)
        output = self.layernorm(output)
        output = self.dropout(output)
        return output


class DisentangledSelfAttention(nn.Cell):
    def __init__(self, config_dict=None, compute_type=mstype.float16, config=None, has_attention_mask=False):
        super().__init__()
        if config:
            self.config = config
        else:
            self.config = ModelConfig.from_dict(config_dict)
        self.has_attention_mask = has_attention_mask
        self.num_attention_heads = self.config.num_attention_heads
        _attention_head_size = int(self.config.hidden_size / self.config.num_attention_heads)
        self.attention_head_size = getattr(self.config, 'attention_head_size', _attention_head_size)
        self.all_head_size = self.num_attention_heads * self.attention_head_size
        initializer_range = self.config.initializer_range
        act_fn = self.config.hidden_act
        weight = TruncatedNormal(initializer_range)

        self.query_proj = nn.Dense(self.config.hidden_size, 
                                    self.all_head_size, 
                                    activation=act_fn, 
                                    weight_init=weight).to_float(compute_type)
        self.key_proj = nn.Dense(self.config.hidden_size, 
                                    self.all_head_size, 
                                    activation=act_fn, 
                                    weight_init=weight).to_float(compute_type)
        self.value_proj = nn.Dense(self.config.hidden_size, 
                                    self.all_head_size, 
                                    activation=act_fn, 
                                    weight_init=weight).to_float(compute_type)

        self.share_att_key = getattr(self.config, 'share_att_key', False)
        self.pos_att_type = [x.strip() for x in getattr(self.config, 'pos_att_type', 'c2p').lower().split('|')] # c2p|p2c
        self.relative_attention = getattr(self.config, 'relative_attention', False)

        if self.relative_attention:
            self.position_buckets = getattr(self.config, 'position_buckets', -1)
            self.max_relative_positions = getattr(self.config, 'max_relative_positions', -1)
            if self.max_relative_positions < 1:
                self.max_relative_positions = self.config.max_position_embeddings
            self.pos_ebd_size = self.max_relative_positions
            if self.position_buckets > 0:
                self.pos_ebd_size = self.position_buckets
                # For backward compitable

            self.pos_dropout = nn.Dropout(self.config.hidden_dropout_prob)

            if (not self.share_att_key):
                if 'c2p' in self.pos_att_type or 'p2p' in self.pos_att_type:
                    self.pos_key_proj = nn.Dense(self.config.hidden_size, 
                                                self.all_head_size, 
                                                activation=act_fn, 
                                                weight_init=weight).to_float(compute_type)
                if 'p2c' in self.pos_att_type or 'p2p' in self.pos_att_type:
                    self.pos_query_proj = nn.Dense(self.config.hidden_size, 
                                                    self.all_head_size, 
                                                    activation=act_fn, 
                                                    weight_init=weight,
                                                    has_bias=False).to_float(compute_type)

        self.dropout = nn.Dropout(self.config.attention_probs_dropout_prob)

        self.shape_return = (-1, self.num_attention_heads * self.attention_head_size )
        self.cast_compute_type = SaturateCast(dst_type=compute_type)
        self.trans_shape = (0, 2, 1, 3)
        self.trans_shape_relative = (2, 0, 1, 3)
        self.trans_shape_position = (1, 2, 0, 3)
        self.multiply_data = -10000.0
        self.scores_mul = 1.0 / math.sqrt(float(self.attention_head_size ))
        self.reshape = P.Reshape()
        self.shape_from_2d = (-1, self.config.hidden_size)
        self.shape_to_2d = (-1, self.config.hidden_size)
        self.matmul_trans_b = P.BatchMatMul(transpose_b=True)
        self.multiply = P.Mul()
        self.transpose = P.Transpose()
        self.matmul = P.BatchMatMul()
        self.softmax = nn.Softmax()
        self.dropout = nn.Dropout(self.config.attention_probs_dropout_prob)
        self.gather = O.GatherD()
        self.depend = P.Depend()

        # if self.has_attention_mask:
        self.expand_dims = P.ExpandDims()
        self.sub = P.Sub()
        self.add = P.Add()
        self.cast = P.Cast()
        self.get_dtype = P.DType()

    def transpose_for_scores(self, x, dense, attention_heads, target_shape, seq_len):
        x = self.reshape(x, target_shape)
        x = dense(x)
        x = self.reshape(x, (-1, seq_len, attention_heads, self.attention_head_size))
        x = self.transpose(x, self.trans_shape).view(-1, x.shape[1], x.shape[-1])
        return x

    def construct(self, hidden_states, query_states, attention_mask, return_att=False, relative_pos=None, rel_embeddings=None, **kwargs):
        if query_states is None:
            query_states = hidden_states
        
        shape_from = F.shape(attention_mask)[2]
        query_states = self.depend(query_states, shape_from)
        query_layer = self.transpose_for_scores(query_states, self.query_proj, self.num_attention_heads, self.shape_from_2d, shape_from)
        key_layer = self.transpose_for_scores(hidden_states, self.key_proj, self.num_attention_heads, self.shape_to_2d, shape_from)
        value_layer = self.transpose_for_scores(hidden_states, self.value_proj, self.num_attention_heads, self.shape_to_2d, shape_from)
       
        rel_att = None
        # Take the dot product between "query" and "key" to get the raw attention scores.
        scale_factor = 1
        if 'c2p' in self.pos_att_type:
            scale_factor += 1
        if 'p2c' in self.pos_att_type:
            scale_factor += 1
        if 'p2p' in self.pos_att_type:
            scale_factor += 1
        # scale = 1 / numpy.sqrt(query_layer.shape[-1] * scale_factor)
        attention_scores = self.matmul(query_layer, key_layer.transpose(0, 2, 1) * self.scores_mul)
        # if self.relative_attention:
        #     rel_embeddings = self.pos_dropout(rel_embeddings)
        #     rel_att = self.disentangled_attention_bias(query_layer, key_layer, relative_pos, rel_embeddings, scale_factor)

        if rel_att is not None:
            attention_scores = (attention_scores + rel_att)
        attention_scores = (attention_scores - attention_scores.max(-1, keepdims=True)).astype(hidden_states.dtype)
        attention_scores = attention_scores.view(-1, self.num_attention_heads, attention_scores.shape[-2], attention_scores.shape[-1])

        # bxhxlxd
        # if self.has_attention_mask:
        attention_mask = self.expand_dims(attention_mask, 1)
        multiply_out = self.sub(self.cast(F.tuple_to_array((1.0,)), self.get_dtype(attention_scores)),
                                self.cast(attention_mask, self.get_dtype(attention_scores)))
        # adder = self.multiply(multiply_out, self.multiply_data)
        attention_scores = self.add(self.multiply(multiply_out, self.multiply_data), attention_scores)
        attention_scores = self.softmax(attention_scores)
        attention_scores = F.cast(self.dropout(attention_scores), value_layer.dtype)
        context_layer = self.matmul(attention_scores.view(-1, attention_scores.shape[-2], attention_scores.shape[-1]), value_layer)
        context_layer = context_layer.view(-1, self.num_attention_heads, context_layer.shape[-2], context_layer.shape[-1]).transpose(self.trans_shape)
        new_context_layer_shape = context_layer.shape[:-2] + (-1,)
        context_layer = context_layer.view(*new_context_layer_shape)

        return context_layer


    def disentangled_attention_bias(self, query_layer, key_layer, relative_pos, rel_embeddings, scale_factor):
        if relative_pos is None:
            q = query_layer.shape[-2]
            relative_pos = build_relative_position(q, key_layer.shape[-2], 
                                                    bucket_size = self.position_buckets, 
                                                    max_position = self.max_relative_positions)
        if relative_pos.ndim == 2:
            relative_pos = relative_pos.view((1, 1, relative_pos.shape[0], relative_pos.shape[1]))
        elif relative_pos.ndim == 3:
            relative_pos = relative_pos.view((relative_pos.shape[0], 1, relative_pos.shape[1], relative_pos.shape[2]))
        # bxhxqxk
        elif relative_pos.ndim != 4:
            raise ValueError(f'Relative postion ids must be of dim 2 or 3 or 4. {relative_pos.ndim}')

        att_span = self.pos_ebd_size

        rel_embeddings = F.expand_dims(rel_embeddings[self.pos_ebd_size - att_span: self.pos_ebd_size + att_span, :], 0)#.repeat(query_layer.shape[0]//self.num_attention_heads, 1, 1)
        if self.share_att_key:
            pos_query_layer = self.transpose_for_scores(rel_embeddings, 
                                                        self.query_proj, 
                                                        self.num_attention_heads, 
                                                        self.shape_to_2d, 
                                                        rel_embeddings.shape[1])
            pos_query_layer = pos_query_layer.repeat(query_layer.shape[0] // self.num_attention_heads, axis=0) 
            pos_key_layer = self.transpose_for_scores(rel_embeddings, 
                                                      self.key_proj, 
                                                      self.num_attention_heads, 
                                                      self.shape_to_2d, 
                                                      rel_embeddings.shape[1])
            pos_key_layer = pos_key_layer.repeat(query_layer.shape[0] // self.num_attention_heads, axis=0) 
            
            # pos_query_layer = self.transpose_for_scores(self.query_proj(rel_embeddings), self.num_attention_heads)\
            #     .repeat(query_layer.shape[0]//self.num_attention_heads, 1, 1) #.split(self.all_head_size, dim=-1)
            # pos_key_layer = self.transpose_for_scores(self.key_proj(rel_embeddings), self.num_attention_heads)\
            #     .repeat(query_layer.shape[0]//self.num_attention_heads, 1, 1) #.split(self.all_head_size, dim=-1)
        else:
            if 'c2p' in self.pos_att_type or 'p2p' in self.pos_att_type:
                pos_key_layer = self.transpose_for_scores(rel_embeddings, 
                                                          self.pos_key_proj, 
                                                          self.num_attention_heads, 
                                                          self.shape_to_2d, 
                                                          rel_embeddings.shape[1])
                pos_key_layer = pos_key_layer.repeat(query_layer.shape[0]//self.num_attention_heads, axis=0) 

                # pos_key_layer = self.transpose_for_scores(self.pos_key_proj(rel_embeddings), self.num_attention_heads)\
                #     .repeat(query_layer.shape[0]//self.num_attention_heads, 1, 1) #.split(self.all_head_size, dim=-1)
            if 'p2c' in self.pos_att_type or 'p2p' in self.pos_att_type:
                pos_query_layer = self.transpose_for_scores(rel_embeddings, 
                                                            self.pos_query_proj, 
                                                            self.num_attention_heads, 
                                                            self.shape_to_2d, 
                                                            rel_embeddings.shape[1])
                pos_query_layer = pos_query_layer.repeat(query_layer.shape[0]//self.num_attention_heads, axis=0) 
                
                # pos_query_layer = self.transpose_for_scores(self.pos_query_proj(rel_embeddings), self.num_attention_heads)\
                #     .repeat(query_layer.shape[0]//self.num_attention_heads, 1, 1) #.split(self.all_head_size, dim=-1)

        score = 0
        # content->position
        if 'c2p' in self.pos_att_type:
            scale = 1 / math.sqrt(pos_key_layer.shape[-1] * scale_factor)
            c2p_att = self.matmul(query_layer, pos_key_layer.transpose(0, 2, 1).astype(query_layer.dtype) * scale)
            c2p_pos = O.clip_by_value(relative_pos + att_span, 0, att_span * 2 - 1)
            index = O.BroadcastTo((query_layer.shape[0], query_layer.shape[1], relative_pos.shape[-1]))(c2p_pos.squeeze(0))
            c2p_att = self.gather(c2p_att, -1, index)
            score += c2p_att

        # position->content
        if 'p2c' in self.pos_att_type or 'p2p' in self.pos_att_type:
            scale = 1 / math.sqrt(pos_query_layer.shape[-1]*scale_factor)
            if key_layer.shape[-2] != query_layer.shape[-2]:
                r_pos = build_relative_position(key_layer.shape[-2], key_layer.shape[-2], 
                                                bucket_size = self.position_buckets, 
                                                max_position = self.max_relative_positions)
                r_pos = F.expand_dims(r_pos, 0)
            else:
                r_pos = relative_pos

            p2c_pos = O.clip_by_value(-r_pos + att_span, 0, att_span * 2 - 1)
            if query_layer.shape[-2] != key_layer.shape[-2]:
                pos_index = F.expand_dims(relative_pos[:, :, :, 0], -1)

        if 'p2c' in self.pos_att_type:
            p2c_att = self.matmul(key_layer, pos_query_layer.transpose(0, 2, 1).astype(key_layer.dtype) * scale)
            index = O.BroadcastTo((query_layer.shape[0], key_layer.shape[-2], key_layer.shape[-2]))(p2c_pos.squeeze(0))
            p2c_att = self.gather(p2c_att, -1, index).transpose(0, 2, 1)
            if query_layer.shape[-2] != key_layer.shape[-2]:
                index = O.BroadcastTo(p2c_att.shape[:2] + (pos_index.shape[-2], key_layer.shape[-2]))(p2c_att)
                p2c_att = self.gather(p2c_att, -2, index)
            score += p2c_att

        # position->position
        if 'p2p' in self.pos_att_type:
            pos_query = pos_query_layer[:, :, att_span:, :]
            p2p_att = self.matmul(pos_query, pos_key_layer.transpose(0, 2, 1))
            p2p_att = O.BroadcastTo(query_layer.shape[:2] + p2p_att.shape[2:])(p2p_att)
            if query_layer.shape[-2] != key_layer.shape[-2]:
                index = O.BroadcastTo(query_layer.shape[:2] + (pos_index.shape[-2], p2p_att.shape[-1]))(pos_index)
                p2p_att = self.gather(p2p_att, -2, index)
            p2p_att = self.gather(p2p_att, -1, O.BroadcastTo((query_layer.shape[0], query_layer.shape[1], query_layer.shape[2], relative_pos.shape[-1])))(c2p_pos)
            score += p2p_att

        return score


class BertAttention(nn.Cell):
    """
    Apply multi-headed attention from "from_tensor" to "to_tensor".

    Args:
        from_tensor_width (int): Size of last dim of from_tensor.
        to_tensor_width (int): Size of last dim of to_tensor.
        num_attention_heads (int): Number of attention heads. Default: 1.
        size_per_head (int): Size of each attention head. Default: 512.
        query_act (str): Activation function for the query transform. Default: None.
        key_act (str): Activation function for the key transform. Default: None.
        value_act (str): Activation function for the value transform. Default: None.
        has_attention_mask (bool): Specifies whether to use attention mask. Default: False.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.0.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        compute_type (:class:`mindspore.dtype`): Compute type in BertAttention. Default: mstype.float32.
    """
    def __init__(self,
                 from_tensor_width,
                 to_tensor_width,
                 num_attention_heads=1,
                 size_per_head=512,
                 query_act=None,
                 key_act=None,
                 value_act=None,
                 has_attention_mask=False,
                 attention_probs_dropout_prob=0.0,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 use_relative_positions=False,
                 compute_type=mstype.float32):

        super(BertAttention, self).__init__()
        self.num_attention_heads = num_attention_heads
        self.size_per_head = size_per_head
        self.has_attention_mask = has_attention_mask
        self.use_relative_positions = use_relative_positions

        self.scores_mul = 1.0 / math.sqrt(float(self.size_per_head))
        self.reshape = P.Reshape()
        self.shape_from_2d = (-1, from_tensor_width)
        self.shape_to_2d = (-1, to_tensor_width)
        weight = TruncatedNormal(initializer_range)
        units = num_attention_heads * size_per_head
        self.query = nn.Dense(from_tensor_width,
                                    units,
                                    activation=query_act,
                                    weight_init=weight).to_float(compute_type)
        self.key = nn.Dense(to_tensor_width,
                                  units,
                                  activation=key_act,
                                  weight_init=weight).to_float(compute_type)
        self.value = nn.Dense(to_tensor_width,
                                    units,
                                    activation=value_act,
                                    weight_init=weight).to_float(compute_type)

        self.matmul_trans_b = P.BatchMatMul(transpose_b=True)
        self.multiply = P.Mul()
        self.transpose = P.Transpose()
        self.trans_shape = (0, 2, 1, 3)
        self.trans_shape_relative = (2, 0, 1, 3)
        self.trans_shape_position = (1, 2, 0, 3)
        self.multiply_data = -10000.0
        self.matmul = P.BatchMatMul()

        self.softmax = nn.Softmax()
        self.dropout = nn.Dropout(attention_probs_dropout_prob)

        if self.has_attention_mask:
            self.expand_dims = P.ExpandDims()
            self.sub = P.Sub()
            self.add = P.Add()
            self.cast = P.Cast()
            self.get_dtype = P.DType()

        self.shape_return = (-1, num_attention_heads * size_per_head)

        self.cast_compute_type = SaturateCast(dst_type=compute_type)
        if self.use_relative_positions:
            self._generate_relative_positions_embeddings = \
                RelaPosEmbeddingsGenerator(depth=size_per_head,
                                           max_relative_position=16,
                                           initializer_range=initializer_range,
                                           use_one_hot_embeddings=use_one_hot_embeddings)

    def construct(self, to_tensor, from_tensor, attention_mask, **kwargs):
        """reshape 2d/3d input tensors to 2d"""
        shape_from = F.shape(attention_mask)[2]
        from_tensor = F.depend(from_tensor, shape_from)
        from_tensor_2d = self.reshape(from_tensor, self.shape_from_2d)
        to_tensor_2d = self.reshape(to_tensor, self.shape_to_2d)
        query_out = self.query(from_tensor_2d)
        key_out = self.key(to_tensor_2d)
        value_out = self.value(to_tensor_2d)

        query_layer = self.reshape(query_out, (-1, shape_from, self.num_attention_heads, self.size_per_head))
        query_layer = self.transpose(query_layer, self.trans_shape)
        key_layer = self.reshape(key_out, (-1, shape_from, self.num_attention_heads, self.size_per_head))
        key_layer = self.transpose(key_layer, self.trans_shape)

        attention_scores = self.matmul_trans_b(query_layer, key_layer)

        # use_relative_position, supplementary logic
        if self.use_relative_positions:
            # relations_keys is [F|T, F|T, H]
            relations_keys = self._generate_relative_positions_embeddings(shape_from)
            relations_keys = self.cast_compute_type(relations_keys)
            # query_layer_t is [F, B, N, H]
            query_layer_t = self.transpose(query_layer, self.trans_shape_relative)
            # query_layer_r is [F, B * N, H]
            query_layer_r = self.reshape(query_layer_t,
                                         (shape_from,
                                          -1,
                                          self.size_per_head))
            # key_position_scores is [F, B * N, F|T]
            key_position_scores = self.matmul_trans_b(query_layer_r,
                                                      relations_keys)
            # key_position_scores_r is [F, B, N, F|T]
            key_position_scores_r = self.reshape(key_position_scores,
                                                 (shape_from,
                                                  -1,
                                                  self.num_attention_heads,
                                                  shape_from))
            # key_position_scores_r_t is [B, N, F, F|T]
            key_position_scores_r_t = self.transpose(key_position_scores_r,
                                                     self.trans_shape_position)
            attention_scores = attention_scores + key_position_scores_r_t

        attention_scores = self.multiply(self.scores_mul, attention_scores)

        if self.has_attention_mask:
            attention_mask = self.expand_dims(attention_mask, 1)
            multiply_out = self.sub(self.cast(F.tuple_to_array((1.0,)), self.get_dtype(attention_scores)),
                                    self.cast(attention_mask, self.get_dtype(attention_scores)))

            adder = self.multiply(multiply_out, self.multiply_data)
            attention_scores = self.add(adder, attention_scores)

        attention_probs = self.softmax(attention_scores)
        attention_probs = self.dropout(attention_probs)

        value_layer = self.reshape(value_out, (-1, shape_from, self.num_attention_heads, self.size_per_head))
        value_layer = self.transpose(value_layer, self.trans_shape)
        context_layer = self.matmul(attention_probs, value_layer)

        # use_relative_position, supplementary logic
        if self.use_relative_positions:
            # relations_values is [F|T, F|T, H]
            relations_values = self._generate_relative_positions_embeddings(shape_from)
            relations_values = self.cast_compute_type(relations_values)
            # attention_probs_t is [F, B, N, T]
            attention_probs_t = self.transpose(attention_probs, self.trans_shape_relative)
            # attention_probs_r is [F, B * N, T]
            attention_probs_r = self.reshape(
                attention_probs_t,
                (shape_from,
                 -1,
                 shape_from))
            # value_position_scores is [F, B * N, H]
            value_position_scores = self.matmul(attention_probs_r,
                                                relations_values)
            # value_position_scores_r is [F, B, N, H]
            value_position_scores_r = self.reshape(value_position_scores,
                                                   (shape_from,
                                                    -1,
                                                    self.num_attention_heads,
                                                    self.size_per_head))
            # value_position_scores_r_t is [B, N, F, H]
            value_position_scores_r_t = self.transpose(value_position_scores_r,
                                                       self.trans_shape_position)
            context_layer = context_layer + value_position_scores_r_t

        context_layer = self.transpose(context_layer, self.trans_shape)
        context_layer = self.reshape(context_layer, context_layer.shape[:2] + (-1,))

        return context_layer


class Attention(nn.Cell):
    """
    Apply self-attention.

    Args:
        hidden_size (int): Size of the bert encoder layers.
        num_attention_heads (int): Number of attention heads. Default: 12.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.1.
        use_one_hot_embeddings (bool): Specifies whether to use one_hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        compute_type (:class:`mindspore.dtype`): Compute type in BertSelfAttention. Default: mstype.float32.
    """
    def __init__(self,
                 hidden_size,
                 attention=BertAttention,
                 num_attention_heads=12,
                 attention_probs_dropout_prob=0.1,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 compute_type=mstype.float32,
                 **kwargs):
        super(Attention, self).__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError("The hidden size (%d) is not a multiple of the number "
                             "of attention heads (%d)" % (hidden_size, num_attention_heads))

        self.size_per_head = int(hidden_size / num_attention_heads)
        # if config is not None:
        #     self.self = DisentangledSelfAttention(config=config, compute_type=compute_type)
        # else:
        #     self.self = BertAttention(
        #         from_tensor_width=hidden_size,
        #         to_tensor_width=hidden_size,
        #         num_attention_heads=num_attention_heads,
        #         size_per_head=self.size_per_head,
        #         attention_probs_dropout_prob=attention_probs_dropout_prob,
        #         use_one_hot_embeddings=use_one_hot_embeddings,
        #         initializer_range=initializer_range,
        #         use_relative_positions=use_relative_positions,
        #         has_attention_mask=True,
        #         compute_type=compute_type)
        if isinstance(attention, type):
            self.attention = attention(
                                from_tensor_width=hidden_size,
                                to_tensor_width=hidden_size,
                                num_attention_heads=num_attention_heads,
                                size_per_head=self.size_per_head,
                                attention_probs_dropout_prob=attention_probs_dropout_prob,
                                use_one_hot_embeddings=use_one_hot_embeddings,
                                initializer_range=initializer_range,
                                use_relative_positions=use_relative_positions,
                                has_attention_mask=True,
                                compute_type=compute_type)
        else:
            self.attention = copy.deepcopy(attention)

        self.output = BertOutput(in_channels=hidden_size,
                                 out_channels=hidden_size,
                                 initializer_range=initializer_range,
                                 dropout_prob=hidden_dropout_prob,
                                 compute_type=compute_type)
        self.reshape = P.Reshape()
        self.shape = (-1, hidden_size)

    def construct(self, input_tensor, query_states, attention_mask, **kwargs):
        attention_output = self.attention(input_tensor, query_states, attention_mask, **kwargs)
        output = self.output(attention_output, input_tensor)
        return output


class BertOutput(nn.Cell):
    """
    Apply a linear computation to hidden status and a residual computation to input.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        dropout_prob (float): The dropout probability. Default: 0.1.
        compute_type (:class:`mindspore.dtype`): Compute type in BertTransformer. Default: mstype.float32.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 initializer_range=0.02,
                 dropout_prob=0.1,
                 compute_type=mstype.float32):
        super(BertOutput, self).__init__()
        self.dense = nn.Dense(in_channels, out_channels,
                              weight_init=TruncatedNormal(initializer_range)).to_float(compute_type)
        self.dropout = nn.Dropout(dropout_prob)
        self.dropout_prob = dropout_prob
        self.add = P.Add()
        self.LayerNorm = nn.LayerNorm((out_channels,)).to_float(compute_type)
        self.cast = P.Cast()

    def construct(self, hidden_status, input_tensor):
        output = self.dense(hidden_status)
        output = self.dropout(output)
        output = self.add(input_tensor, output)
        output = self.LayerNorm(output)
        return output


class RelaPosMatrixGenerator(nn.Cell):
    """
    Generates matrix of relative positions between inputs.

    Args:
        length (int): Length of one dim for the matrix to be generated.
        max_relative_position (int): Max value of relative position.
    """
    def __init__(self, max_relative_position):
        super(RelaPosMatrixGenerator, self).__init__()
        self._max_relative_position = max_relative_position
        self._min_relative_position = -max_relative_position

        self.tile = P.Tile()
        self.range_mat = P.Reshape()
        self.sub = P.Sub()
        self.expanddims = P.ExpandDims()
        self.cast = P.Cast()

    def construct(self, length):
        """Generates matrix of relative positions between inputs."""
        range_vec_row_out = self.cast(F.tuple_to_array(F.make_range(length)), mstype.int32)
        range_vec_col_out = self.range_mat(range_vec_row_out, (length, -1))
        tile_row_out = self.tile(range_vec_row_out, (length,))
        tile_col_out = self.tile(range_vec_col_out, (1, length))
        range_mat_out = self.range_mat(tile_row_out, (length, length))
        transpose_out = self.range_mat(tile_col_out, (length, length))
        distance_mat = self.sub(range_mat_out, transpose_out)

        distance_mat_clipped = C.clip_by_value(distance_mat,
                                               self._min_relative_position,
                                               self._max_relative_position)

        # Shift values to be >=0. Each integer still uniquely identifies a
        # relative position difference.
        final_mat = distance_mat_clipped + self._max_relative_position
        return final_mat


class RelaPosEmbeddingsGenerator(nn.Cell):
    """
    Generates tensor of size [length, length, depth].

    Args:
        length (int): Length of one dim for the matrix to be generated.
        depth (int): Size of each attention head.
        max_relative_position (int): Maxmum value of relative position.
        initializer_range (float): Initialization value of TruncatedNormal.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
    """
    def __init__(self,
                 depth,
                 max_relative_position,
                 initializer_range,
                 use_one_hot_embeddings=False):
        super(RelaPosEmbeddingsGenerator, self).__init__()
        self.depth = depth
        self.vocab_size = max_relative_position * 2 + 1
        self.use_one_hot_embeddings = use_one_hot_embeddings

        self.embeddings_table = Parameter(
            initializer(TruncatedNormal(initializer_range),
                        [self.vocab_size, self.depth]))

        self.relative_positions_matrix = RelaPosMatrixGenerator(max_relative_position=max_relative_position)
        self.reshape = P.Reshape()
        self.one_hot = nn.OneHot(depth=self.vocab_size)
        self.shape = P.Shape()
        self.gather = P.Gather()  # index_select
        self.matmul = P.BatchMatMul()

    def construct(self, length):
        """Generate embedding for each relative position of dimension depth."""
        relative_positions_matrix_out = self.relative_positions_matrix(length)

        if self.use_one_hot_embeddings:
            flat_relative_positions_matrix = self.reshape(relative_positions_matrix_out, (-1,))
            one_hot_relative_positions_matrix = self.one_hot(
                flat_relative_positions_matrix)
            embeddings = self.matmul(one_hot_relative_positions_matrix, self.embeddings_table)
            my_shape = self.shape(relative_positions_matrix_out) + (self.depth,)
            embeddings = self.reshape(embeddings, my_shape)
        else:
            embeddings = self.gather(self.embeddings_table,
                                     relative_positions_matrix_out, 0)
        return embeddings


class EncoderLayer(nn.Cell):
    """
    Encoder cells used in BertTransformer.

    Args:
        hidden_size (int): Size of the bert encoder layers. Default: 768.
        num_attention_heads (int): Number of attention heads. Default: 12.
        intermediate_size (int): Size of intermediate layer. Default: 3072.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.02.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        hidden_act (str): Activation function. Default: "gelu".
        compute_type (:class:`mindspore.dtype`): Compute type in attention. Default: mstype.float32.
    """
    def __init__(self,
                 hidden_size=768,
                 num_attention_heads=12,
                 intermediate_size=3072,
                 attention_probs_dropout_prob=0.02,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 hidden_act="gelu",
                 attension=BertAttention,
                 compute_type=mstype.float32):
        super(EncoderLayer, self).__init__()
        self.attention = Attention(
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=initializer_range,
            hidden_dropout_prob=hidden_dropout_prob,
            use_relative_positions=use_relative_positions,
            attention=attension,
            compute_type=compute_type)
        self.intermediate = nn.Dense(in_channels=hidden_size,
                                     out_channels=intermediate_size,
                                     activation=hidden_act,
                                     weight_init=TruncatedNormal(initializer_range)).to_float(compute_type)
        self.output = BertOutput(in_channels=intermediate_size,
                                 out_channels=hidden_size,
                                 initializer_range=initializer_range,
                                 dropout_prob=hidden_dropout_prob,
                                 compute_type=compute_type)

    def construct(self, hidden_states, query_states, attention_mask, **kwargs):
        # self-attention
        attention_output = self.attention(hidden_states, query_states, attention_mask, **kwargs)
        # feed construct
        intermediate_output = self.intermediate(attention_output)
        # add and normalize
        output = self.output(intermediate_output, attention_output)
        return output


class ConvLayer(nn.Cell):
    def __init__(self, conv_kernel_size=3, conv_groups=1, conv_act='tanh', 
                 hidden_size=768, hidden_dropout_prob=0.1, compute_type=mstype.float32, **kwargs):
        super().__init__()
        self.conv_act = nn.get_activation(conv_act)
        self.conv = nn.Conv1d(hidden_size, hidden_size, conv_kernel_size, group=conv_groups).to_float(compute_type)
        self.LayerNorm = nn.LayerNorm((hidden_size,), epsilon=kwargs.get('layer_norm_eps', 1e-7))
        self.dropout = nn.Dropout(hidden_dropout_prob)

    def construct(self, hidden_states, residual_states, input_mask):
        out = self.conv(hidden_states.transpose(0, 2, 1)).transpose(0, 2, 1)
        rmask = 1 - input_mask
        out = P.MaskedFill()(out, O.BroadcastTo(out.shape)(F.expand_dims(rmask, -1)).astype(mstype.bool_), 0.0)
        out = self.conv_act(self.dropout(out))
        output_states = self.LayerNorm(residual_states + out)
        return output_states


class SaturateCast(nn.Cell):
    """
    Performs a safe saturating cast. This operation applies proper clamping before casting to prevent
    the danger that the value will overflow or underflow.

    Args:
        src_type (:class:`mindspore.dtype`): The type of the elements of the input tensor. Default: mstype.float32.
        dst_type (:class:`mindspore.dtype`): The type of the elements of the output tensor. Default: mstype.float32.
    """
    def __init__(self, src_type=mstype.float32, dst_type=mstype.float32):
        super(SaturateCast, self).__init__()
        np_type = mstype.dtype_to_nptype(dst_type)

        self.tensor_min_type = float(numpy.finfo(np_type).min)
        self.tensor_max_type = float(numpy.finfo(np_type).max)

        self.min_op = P.Minimum()
        self.max_op = P.Maximum()
        self.cast = P.Cast()
        self.dst_type = dst_type

    def construct(self, x):
        out = self.max_op(x, self.tensor_min_type)
        out = self.min_op(out, self.tensor_max_type)
        return self.cast(out, self.dst_type)


class Encoder(nn.Cell):
    """
    Multi-layer bert transformer.

    Args:
        hidden_size (int): Size of the encoder layers.
        num_hidden_layers (int): Number of hidden layers in encoder cells.
        num_attention_heads (int): Number of attention heads in encoder cells. Default: 12.
        intermediate_size (int): Size of intermediate layer in encoder cells. Default: 3072.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.1.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        hidden_act (str): Activation function used in the encoder cells. Default: "gelu".
        compute_type (:class:`mindspore.dtype`): Compute type in BertTransformer. Default: mstype.float32.
        return_all_encoders (bool): Specifies whether to return all encoders. Default: False.
    """
    def __init__(self,
                 hidden_size,
                 num_hidden_layers,
                 num_attention_heads=12,
                 intermediate_size=3072,
                 attention_probs_dropout_prob=0.1,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 hidden_act="gelu",
                 attention=BertAttention,
                 compute_type=mstype.float32,
                 return_all_encoders=False,
                 max_position_embeddings=512,
                 **kwargs):
        super(Encoder, self).__init__()
        self.return_all_encoders = return_all_encoders
        if 'config' in kwargs:
            kwargs.update(kwargs['config'].to_dict())
        layers = []
        self.num_hidden_layers = num_hidden_layers
        for _ in range(num_hidden_layers):
            layer = EncoderLayer(hidden_size=hidden_size,
                                 num_attention_heads=num_attention_heads,
                                 intermediate_size=intermediate_size,
                                 attention_probs_dropout_prob=attention_probs_dropout_prob,
                                 use_one_hot_embeddings=use_one_hot_embeddings,
                                 initializer_range=initializer_range,
                                 hidden_dropout_prob=hidden_dropout_prob,
                                 use_relative_positions=use_relative_positions,
                                 hidden_act=hidden_act,
                                 attension=attention,
                                 compute_type=compute_type)
            layers.append(layer)

        self.layer = nn.CellList(layers)
        self.relative_attention = use_relative_positions
        if use_relative_positions:
            self.position_buckets = kwargs.get('position_buckets', -1)
            self.max_relative_positions = kwargs.get('max_relative_positions', -1)
            if self.max_relative_positions < 1:
                    self.max_relative_positions = max_position_embeddings
            self.create_rel_embeddings(self.max_relative_positions, position_buckets=self.position_buckets,
                                        hidden_size=hidden_size)
        else:
            self.rel_embeddings = None

        self.norm_rel_ebd = [x.strip() for x in kwargs.get('norm_rel_ebd', 'none').lower().split('|')]
        if 'layer_norm' in self.norm_rel_ebd:
            self.LayerNorm = nn.LayerNorm((hidden_size,), epsilon=kwargs.get('layer_norm_eps', 1e-7))
        else:
            self.LayerNorm = None
        
        # add 1-D conv to improve aggregation
        kernel_size = kwargs.get('conv_kernel_size', 0)
        self.with_conv = False
        if kernel_size > 0:
            self.with_conv = True
            self.conv = ConvLayer(conv_kernel_size=kernel_size, hidden_size=hidden_size, 
                                  hidden_dropout_prob=hidden_dropout_prob, compute_type=compute_type)
        self.reshape = P.Reshape()
        self.shape = (-1, hidden_size)


    def create_rel_embeddings(self, max_relative_positions, position_buckets=512, hidden_size=768, **kwargs):
        pos_ebd_size = max_relative_positions * 2
        if position_buckets > 0:
                pos_ebd_size = position_buckets * 2
        self.rel_embeddings = nn.Embedding(pos_ebd_size, hidden_size)

    def construct(self, hidden_states, attention_mask, output_all_encoded_layers=True, 
                    return_att=False, query_states = None, relative_pos=None):
        """Multi-layer bert transformer."""
        # prev_output = self.reshape(hidden_states, self.shape)
        all_encoder_layers = []
        if attention_mask.ndim <= 2:
            input_mask = attention_mask
        else:
            input_mask = (attention_mask.sum(-2) > 0)
        if self.relative_attention:
            relative_pos = get_rel_pos(hidden_states, query_states, relative_pos, 
                                        relative_attention=self.relative_attention,
                                        position_buckets=self.position_buckets,
                                        max_relative_positions=self.max_relative_positions)
        # all_encoder_layers = []
        att_matrices = []
        if isinstance(hidden_states, Tensor):
            next_kv = hidden_states
        else:
            next_kv = hidden_states[0]
            
        if self.relative_attention is not None:
            rel_embeddings = get_rel_embedding(self.rel_embeddings.embedding_table, self.relative_attention, self.norm_rel_ebd, self.LayerNorm)
        else:
            rel_embeddings = None

        output_states = hidden_states
        att_m = None
        for i, layer_module in enumerate(self.layer):
            output_states = layer_module(next_kv, query_states, attention_mask, return_att=return_att, 
                                         relative_pos=relative_pos, rel_embeddings=rel_embeddings)

            if return_att:
                output_states, att_m = output_states
            else:
                att_m = None

            if i == 0 and self.with_conv:
                prenorm = output_states #output['prenorm_states']
                output_states = self.conv(hidden_states, prenorm, input_mask)

            if query_states is not None:
                query_states = output_states
                if isinstance(hidden_states, Sequence):
                    next_kv = hidden_states[i + 1] if i + 1 < len(self.layer) else None
            else:
                next_kv = output_states

            if output_all_encoded_layers:
                all_encoder_layers.append(output_states)
                if return_att:
                    att_matrices.append(att_m)
            else:
                if i == self.num_hidden_layers - 1:
                    all_encoder_layers.append(output_states)
                if return_att:
                    att_matrices.append(att_m)
        # if not output_all_encoded_layers:
        #     all_encoder_layers.append(output_states)
        #     if return_att:
        #         att_matrices.append(att_m)
        return {
            'hidden_states': all_encoder_layers,
            'attention_matrices': att_matrices
            }


        #     layer_output = layer_module(prev_output, attention_mask)
        #     prev_output = layer_output

        #     if self.return_all_encoders:
        #         shape = F.shape(hidden_states)
        #         layer_output = self.reshape(layer_output, shape)
        #         all_encoder_layers = all_encoder_layers + (layer_output,)

        # if not self.return_all_encoders:
        #     shape = F.shape(hidden_states)
        #     prev_output = self.reshape(prev_output, shape)
        #     all_encoder_layers = all_encoder_layers + (prev_output,)
        # return all_encoder_layers


class Pooler(nn.Cell):
    def __init__(self, hidden_size, initializer_range, compute_type='float16') -> None:
        super().__init__()
        weight_init = TruncatedNormal(initializer_range)
        self.dense = nn.Dense(hidden_size,
                              hidden_size,
                              weight_init=weight_init,
                              activation="tanh").to_float(compute_type)

    def construct(self, hidden_states):
        return self.dense(hidden_states)

class GetMaskedLMOutput(nn.Cell):
    """
    Get masked lm output.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        Tensor, masked lm output.
    """

    def __init__(self, **config):
        super(GetMaskedLMOutput, self).__init__()
        self.width = config.get('hidden_size')
        self.reshape = P.Reshape()
        self.gather = P.Gather()

        weight_init = TruncatedNormal(config.get('initializer_range'))
        self.dense = nn.Dense(self.width,
                              config.get('hidden_size'),
                              weight_init=weight_init,
                              activation=config.get('hidden_act')).to_float(config.get('compute_type'))
        self.LayerNorm = nn.LayerNorm((config.get('hidden_size'),)).to_float(config.get('compute_type'))
        self.bias = Parameter(
            initializer(
                'zero',
                config.get('vocab_size')))
        self.matmul = P.MatMul(transpose_b=True)
        self.log_softmax = nn.LogSoftmax(axis=-1)
        self.shape_flat_offsets = (-1, 1)
        self.last_idx = (-1,)
        self.shape_flat_sequence_tensor = (-1, self.width)
        self.cast = P.Cast()
        self.compute_type = config.get('compute_type')
        self.dtype = config.get('compute_type')

    def construct(self,
                  input_tensor,
                  output_weights,
                  positions):
        """Get output log_probs"""
        input_shape = P.Shape()(input_tensor)
        rng = F.tuple_to_array(F.make_range(input_shape[0]))
        flat_offsets = self.reshape(rng * input_shape[1], self.shape_flat_offsets)
        flat_position = self.reshape(positions + flat_offsets, self.last_idx).astype(mstype.int32)
        flat_sequence_tensor = self.reshape(input_tensor, self.shape_flat_sequence_tensor)
        input_tensor = self.gather(flat_sequence_tensor, flat_position, 0)
        input_tensor = self.cast(input_tensor, self.compute_type)
        output_weights = self.cast(output_weights, self.compute_type)
        input_tensor = self.dense(input_tensor)
        input_tensor = self.LayerNorm(input_tensor)
        logits = self.matmul(input_tensor, output_weights)
        logits = self.cast(logits, self.dtype)
        logits = logits + self.bias
        log_probs = self.log_softmax(logits)
        return log_probs


class GetNextSentenceOutput(nn.Cell):
    """
    Get next sentence output.

    Args:
        config (BertConfig): The config of Bert.

    Returns:
        Tensor, next sentence output.
    """

    def __init__(self, **config):
        super(GetNextSentenceOutput, self).__init__()
        self.log_softmax = P.LogSoftmax()
        weight_init = TruncatedNormal(config.get('initializer_range'))
        self.dense = nn.Dense(config.get('hidden_size'), 2,
                              weight_init=weight_init, has_bias=True).to_float(config.get('compute_type'))
        self.dtype = config.get('compute_type')
        self.cast = P.Cast()

    def construct(self, input_tensor):
        logits = self.dense(input_tensor)
        logits = self.cast(logits, self.dtype)
        log_prob = self.log_softmax(logits)
        return log_prob


class FakeHead(nn.Cell):
    """
    A fake head which return the last arg

    Args:
        None

    Returns:
        Tensor, last arg.
    """

    def __init__(self):
        super(FakeHead, self).__init__()

    def construct(self, *input):
        return input[-1]


class BertEvalHead(nn.Cell):
    """
    Provide bert pre-training evaluation head.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        tuple: Tensor, total loss. Tensor, ppl
    """

    def __init__(self, vocab_size):
        super(BertEvalHead, self).__init__()
        self.vocab_size = vocab_size
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, mstype.float32)
        self.off_value = Tensor(0.0, mstype.float32)
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.reshape = P.Reshape()
        self.last_idx = (-1,)
        self.neg = P.Neg()
        self.cast = P.Cast()
        self.argmax = P.Argmax()

    def construct(self, *sample):
        """Defines the computation performed.
            sample: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                     "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"] + 
                    [prediction_scores, seq_relationship_score]
        """
        input_ids = sample[0]
        input_mask = sample[1]
        token_type_id = sample[2]
        masked_lm_positions = sample[4]
        masked_lm_ids = sample[5]
        masked_lm_weights = sample[6]
        prediction_scores = sample[7]
        next_sentence_labels = sample[3]
        seq_relationship_score = sample[8]
        label_ids = self.reshape(masked_lm_ids, self.last_idx)
        label_weights = self.cast(self.reshape(masked_lm_weights, self.last_idx), mstype.float32)
        one_hot_labels = self.onehot(label_ids, self.vocab_size, self.on_value, self.off_value)

        per_example_loss = self.neg(self.reduce_sum(prediction_scores * one_hot_labels, self.last_idx))
        numerator = self.reduce_sum(label_weights * per_example_loss, ())
        denominator = self.reduce_sum(label_weights, ()) + self.cast(F.tuple_to_array((1e-5,)), mstype.float32)
        masked_lm_loss = numerator / denominator

        # next_sentence_loss
        labels = self.reshape(next_sentence_labels, self.last_idx)
        if labels.max() >= 0:
            one_hot_labels = self.onehot(labels, 2, self.on_value, self.off_value)
            per_example_loss = self.neg(self.reduce_sum(
                one_hot_labels * seq_relationship_score, self.last_idx))
            next_sentence_loss = self.reduce_mean(per_example_loss, self.last_idx)
        else:
            next_sentence_loss = 0

        # total_loss
        total_loss = masked_lm_loss + next_sentence_loss

        bs, _ = F.shape(input_ids)

        index = self.argmax(prediction_scores)
        index = self.reshape(index, (bs, -1))
        eval_acc = F.equal(index, masked_lm_ids)
        eval_acc = self.cast(eval_acc, mstype.float32)
        real_acc = eval_acc * masked_lm_weights
        acc = real_acc.sum()
        total = masked_lm_weights.astype(mstype.float32).sum()

        return total_loss, acc, total, prediction_scores, masked_lm_ids


class ClsEvalHead(nn.Cell):
    """
    Provide bert pre-training evaluation head.

    Args:
        config (BertConfig): The config of BertModel.

    Returns:
        tuple: Tensor, total loss. Tensor, ppl
    """

    def __init__(self, num_labels, return_all=False, fp16=False, dropout=0.1):
        super(ClsEvalHead, self).__init__()
        self.loss_fn = nn.SoftmaxCrossEntropyWithLogits(sparse=True, reduction='mean')

    def construct(self, logits, labels):
        """Defines the computation performed.
        """
        loss = self.loss_fn(logits, labels)
        return (loss, logits, F.cast(labels, mstype.int32))

class EnhancedMaskDecoder(nn.Cell):
  def __init__(self, **kwargs):
    super().__init__()
    self.config = kwargs
    self.position_biased_input = self.config.get('position_biased_input', True)
    # self.lm_head = GetMaskedLMOutput(**kwargs)
    self.loss_fn = nn.SoftmaxCrossEntropyWithLogits(sparse=True)
    self._create_attention_mask_from_input_mask = CreateAttentionMaskFromInputMask()

  def construct(self, ctx_layers, z_states, attention_mask, 
                encoder_layer, embedding_table, relative_attention, 
                norm_rel_ebd, LayerNorm_fn, relative_pos=None):
    # if attention_mask.ndim != 2:
    #   attention_mask = (attention_mask.sum(-2) > 0)
    #   attention_mask = extended_attention_mask * F.expand_dims(extended_attention_mask.squeeze(-2), -1)
    # elif attention_mask.ndim == 3:
    #   attention_mask = F.expand_dims(attention_mask, 1)
    attention_mask = self._create_attention_mask_from_input_mask(attention_mask)
    hidden_states = ctx_layers[-2]
    hidden_states = F.cast(hidden_states, z_states.dtype)
    # layers = [encoder.layer[-1] for _ in range(2)]
    layer = encoder_layer
    if not self.position_biased_input: 
    #   layers = [encoder.layer[-1] for _ in range(2)]
      
      query_states = F.add(z_states, hidden_states) # .astype()
      #   query_states = z_states
      query_mask = attention_mask
      mlm_ctx_layers = []
      rel_embeddings = get_rel_embedding(embedding_table, relative_attention, norm_rel_ebd, LayerNorm_fn)
      #layer_module(next_kv, query_states, attention_mask, return_att=return_att, relative_pos=relative_pos, rel_embeddings=rel_embeddings)
      for i in range(2):
        # TODO: pass relative pos ids
        output = layer(hidden_states, query_states, query_mask, return_att=False, relative_pos=relative_pos, rel_embeddings=rel_embeddings)
        query_states = output
        mlm_ctx_layers.append(query_states)
    else:
      mlm_ctx_layers = [ctx_layers[-1]]
    # mlm_ctx_layers = self.emd_context_layer(ctx_layers, z_states, attention_mask, encoder, relative_pos=relative_pos)
    ctx_layer = mlm_ctx_layers[-1]
    # lm_logits = self.lm_head(ctx_layer, ebd_weight, masked_lm_positions)
    return ctx_layer

  def emd_context_layer(self, encoder_layers, z_states, attention_mask, encoder, relative_pos=None):
    if attention_mask.ndim <= 2:
      extended_attention_mask = attention_mask.view(attention_mask.shape + (1, 1))
      attention_mask = extended_attention_mask * F.expand_dims(extended_attention_mask.squeeze(-2), -1)
    # elif attention_mask.ndim == 3:
    #   attention_mask = F.expand_dims(attention_mask, 1)
    hidden_states = encoder_layers[-2]
    hidden_states = F.cast(hidden_states, z_states.dtype)
    if not self.position_biased_input: 
      layers = [encoder.layer[-1] for _ in range(2)]
      
      query_states = F.add(z_states, hidden_states) # .astype()
      #   query_states = z_states
      query_mask = attention_mask
      outputs = []
      rel_embeddings = get_rel_embedding(encoder.rel_embeddings, encoder.relative_attention, encoder.norm_rel_ebd, encoder.LayerNorm)
      #layer_module(next_kv, query_states, attention_mask, return_att=return_att, relative_pos=relative_pos, rel_embeddings=rel_embeddings)
      for layer in layers:
        # TODO: pass relative pos ids
        output = layer(hidden_states, query_states, query_mask, return_att=False, relative_pos=relative_pos, rel_embeddings=rel_embeddings)
        query_states = output
        outputs.append(query_states)
    else:
      outputs = [encoder_layers[-1]]
    
    return outputs


def make_log_bucket_position(relative_pos, bucket_size, max_position):
    sign = P.Sign()(relative_pos)
    mid = bucket_size // 2
    a = (relative_pos < mid).astype(mstype.int16)
    b = (relative_pos > -mid).astype(mstype.int16)
    abs_pos = np.where(np.bitwise_and(a, b), mid - 1, np.abs(relative_pos)).astype(mstype.float16)
    log_pos = np.ceil(np.log(abs_pos / mid) / np.log((max_position - 1) / mid) * (mid - 1)) + mid
    bucket_pos = np.where(abs_pos <= mid, relative_pos, log_pos * sign).astype(np.int64)
    return bucket_pos

# @lru_cache(maxsize=128)
def build_relative_position(query_size, key_size, bucket_size=-1, max_position=-1):
    q_ids = np.arange(0, query_size)
    k_ids = np.arange(0, key_size)
    rel_pos_ids = q_ids[:, None] - np.tile(k_ids, (q_ids.shape[0],1))
    if bucket_size > 0 and max_position > 0:
        rel_pos_ids = make_log_bucket_position(rel_pos_ids, bucket_size, max_position)
    # rel_pos_ids = Tensor(rel_pos_ids, dtype=mstype.int64)
    rel_pos_ids = rel_pos_ids[:query_size, :]
    rel_pos_ids = F.expand_dims(rel_pos_ids, 0)
    return rel_pos_ids

def get_rel_embedding(rel_embeddings, relative_attention, norm_rel_ebd, layer_norm):
    if relative_attention:
        if 'layer_norm' in norm_rel_ebd:
            rel_embeddings = layer_norm(rel_embeddings)
    return rel_embeddings


def get_rel_pos(hidden_states, query_states=None, relative_pos=None, relative_attention=None, position_buckets=None, max_relative_positions=512):
    if relative_attention and relative_pos is None:
        q = query_states.shape[-2] if query_states is not None else hidden_states.shape[-2]
        relative_pos = build_relative_position(q, hidden_states.shape[-2], bucket_size=position_buckets, max_position=max_relative_positions)
    return relative_pos



# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import Sequence, Mapping
import math
from mindspore.train.callback import Callback

# def batch_apply(batch, fn):
#   if isinstance(batch, torch.Tensor):
#     return fn(batch)
#   elif isinstance(batch, Sequence):
#     return [batch_apply(x, fn) for x in batch]
#   elif isinstance(batch, Mapping):
#     return {x:batch_apply(batch[x], fn) for x in batch}
#   else:
#     raise NotImplementedError(f'Type of {type(batch)} are not supported in batch_apply')

# def batch_to(batch, device):
#   return batch_apply(batch, lambda x: x.to(device))

class LossCallBack(Callback):
    """
    Monitor the loss in training.
    If the loss in NAN or INF terminating training.
    Note:
        if per_print_times is 0 do not print loss.
    Args:
        per_print_times (int): Print loss every times. Default: 1.
    """
    def __init__(self, dataset_size=-1):
        super(LossCallBack, self).__init__()
        self._dataset_size = dataset_size
    def step_end(self, run_context):
        """
        Print loss after each step
        """
        cb_params = run_context.original_args()
        if self._dataset_size > 0:
            percent, epoch_num = math.modf(cb_params.cur_step_num / self._dataset_size)
            if percent == 0:
                percent = 1
                epoch_num -= 1
            print("epoch: {}, current epoch percent: {}, step: {}, outputs are {}"
                  .format(int(epoch_num), "%.3f" % percent, cb_params.cur_step_num, str(cb_params.net_outputs)),
                  flush=True)
        else:
            print("epoch: {}, step: {}, outputs are {}".format(cb_params.cur_epoch_num, cb_params.cur_step_num,
                                                               str(cb_params.net_outputs)), flush=True)


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .models import *
from .config import ModelConfig

models = {
    'bert': BertModel,
    'bert_pretraining': BertPreTraining,
    'deberta': Deberta,
    'deberta_pretraining': DebertaPreTraining
}


def build_transformer_model(
    config_path=None,
    model='bert',
    application='encoder',
    **kwargs
):
    """根据配置文件构建模型，可选加载checkpoint权重
    """
    compute_type = kwargs.get('compute_type', mstype.float16)
    if config_path is not None:
        configs = ModelConfig.from_json_file(config_path)
    else:
        configs = ModelConfig()
    configs.update(**kwargs)
    # mindspore中dropout是keep rate
    if configs['hidden_dropout_prob'] < 0.5:
        configs['hidden_dropout_prob'] = 1 - configs.get('hidden_dropout_prob')
        configs['attention_probs_dropout_prob'] = 1 - configs.get('attention_probs_dropout_prob')
    if 'max_position' not in configs:
        configs['max_position'] = configs.get('max_position_embeddings', 512)
    if 'dropout_rate' not in configs:
        configs['dropout_rate'] = configs.get('hidden_dropout_prob')
    if 'attention_dropout_rate' not in configs:
        configs['attention_dropout_rate'] = configs.get(
            'attention_probs_dropout_prob'
        )
    if 'segment_vocab_size' not in configs:
        configs['segment_vocab_size'] = configs.get('type_vocab_size', 2)

    if 'compute_type' not in configs:
        configs['compute_type'] = compute_type

    if isinstance(model, str):
        model = model.lower()
        MODEL = models[model]
    else:
        MODEL = model

    # application = application.lower()
    # if application in ['lm', 'unilm'] and model in ['electra', 't5']:
    #     raise ValueError(
    #         '"%s" model can not be used as "%s" application.\n' %
    #         (model, application)
    #     )

    # if application == 'lm':
    #     MODEL = extend_with_language_model(MODEL)
    # elif application == 'unilm':
    #     MODEL = extend_with_unified_language_model(MODEL)
    transformer = MODEL(**configs.to_dict(), config=configs)

    # if checkpoint_path is not None:
    #     transformer.load_weights_from_checkpoint(checkpoint_path)

    return transformer, configs


# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================


from queue import Queue,Empty
from threading import Thread
class AsyncDataLoader(object):
  def __init__(self, dataloader, buffer_size=100):
    self.buffer_size = buffer_size
    self.dataloader = dataloader

  def __iter__(self):
    queue = Queue(self.buffer_size)
    dl=iter(self.dataloader)
    def _worker():
      while True:
        try:
          queue.put(next(dl))
        except StopIteration:
          break
      queue.put(None)
    t=Thread(target=_worker)
    t.start()
    while True:
      d = queue.get()
      if d is None:
        break
      yield d
    del t
    del queue

  def __len__(self):
    return len(self.dataloader)




# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from collections import OrderedDict

import numpy as np
import os
import random
import math
from bisect import bisect
import ujson as json
from shutil import copyfile
# from transformers import AutoModelForMultipleChoice

from loguru import logger

from MindsporeTrainer.data import ExampleInstance, ExampleSet, _truncate_segments
from MindsporeTrainer.data.example import *
from MindsporeTrainer.task import EvalData, Task
from MindsporeTrainer.task  import register_task
from utils.metrics import *

import mindspore.nn as nn
import mindspore.ops as ops
import mindspore.dataset as ds
import mindspore.dataset.vision as vision
import mindspore.dataset.transforms as transforms
from mindspore.communication import init, get_rank, get_group_size
from mindspore.nn import Momentum, SoftmaxCrossEntropyWithLogits
import mindspore as ms

from MindsporeTrainer.modeling.modeling_adapter import NetworkWithLoss
from MindsporeTrainer.utils.metrics import BertMetric
from MindsporeTrainer.utils.masker import NGramMaskGenerator
from MindsporeTrainer.data.dynamic_dataset import create_dynamic_dataset
from MindsporeTrainer.modeling.tokenizers import BertTokenizer


@register_task(name="BERT", desc="Basic BERT task")
class BERTTask(Task):
  def __init__(self, data_dir, args=None, **kwargs):
    super().__init__(args, **kwargs)
    self.data_dir = data_dir
    self.args = args
    # self.model_class = resnet50(32, 100)
    self.metric = 'bert'
    self.main_metric = 'perplexity'
    self.optimizer_name = 'Lamb'
    self.tokenizer = BertTokenizer(self.args.vocab_path)
    if args.distribute:
      self.rank_id = get_rank()
      self.rank_size = get_group_size()
    else:
      self.rank_id = 0
      self.rank_size = 1

  def train_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data_path = os.path.join(self.data_dir, 'daizhige.pkl')
    data = self.load_data(data_path, 'GW', max_seq_len)
    # data = ExampleSet(data)
    output_columns = ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                  "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
    return create_dynamic_dataset(data, self.get_feature_fn(), batch_size=batch_size,
                                  output_columns=output_columns, repeat=self.args.num_train_epochs,
                                  num_workers=self.args.data_workers,num_shards=self.rank_size, shard_id=self.rank_id)

  def eval_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data_path = os.path.join(self.data_dir, 'eval.txt')
    data = self.load_txt_data(data_path, 'GW', max_seq_len)

    output_columns = ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                  "masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
    return create_dynamic_dataset(data, self.get_feature_fn(), batch_size=batch_size,
                                  output_columns=output_columns, repeat=1,
                                  num_workers=self.args.data_workers, num_shards=self.rank_size, 
                                  shard_id=self.rank_id)

  def test_data(self, max_seq_len=512, batch_size=32, **kwargs):
    data_path = os.path.join(self.data_dir, 'daizhige.pkl')
    return self.create_dataset(data_path=data_path, max_seq_len=max_seq_len, batch_size=batch_size)

  def _data(self, name, path, type_name = 'dev', ignore_metric=False, max_examples=None, shuffle=False, max_seq_len=512):
    input_src = os.path.join(self.data_dir, path)
    assert os.path.exists(input_src), f"{input_src} doesn't exists"
    data = self.create_dataset(name)

    predict_fn = self.get_predict_fn(data)
    metrics_fn = self.get_metrics()
    return EvalData(name, data,
      metrics_fn = metrics_fn, predict_fn = predict_fn, ignore_metric=ignore_metric, critial_metrics=['accuracy'])

  def get_metrics(self, **kwargs):
    """Calcuate metrics based on prediction results"""
    return OrderedDict(
      bert = BertMetric(self.args.eval_batch_size),
      )

  # def get_predict_fn(self, data):
  #   """Calcuate metrics based on prediction results"""
  #   def predict_fn(logits, output_dir, name, prefix, targets=None):
  #     output=os.path.join(output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
  #     preds = np.argmax(logits, axis=-1)
  #     labels = self.get_labels()
  #     with open(output, 'w', encoding='utf-8') as fs:
  #       fs.write('targets\tpredictions\n')
  #       if targets is not None:
  #         for i,(e, p) in enumerate(zip(targets, preds)):
  #           fs.write(f'{labels[e]}\t{labels[p]}\n')
  #       else:
  #         for i,(e,p) in enumerate(zip(data, preds)):
  #           fs.write(f'{labels[e.label]}\t{labels[p]}\n')
  #   return predict_fn

  def get_eval_fn(self, **kwargs):
    data = kwargs.get('data')
    if data is None:
      data = self.eval_data(**kwargs)
    def run_eval(model, name, prefix):
      '''
      args: 
        model: Model instance
        name: evaluate name
        prefix: prefix of file
      return:
        float, main metric of this task, used to save best metric model
      '''
      # from wywLM.utils.callbacks import EvalResultsCallback, LossMoniter
      # rc = EvalResultsCallback(result_fn='argmax')
      res = model.eval(data, dataset_sink_mode=False)
      res = res['bert']
      main_metric = res[self.main_metric]
      if self.rank_id == 0:
        output=os.path.join(self.args.output_dir, 'submit-{}-{}.tsv'.format(name, prefix))
        metric_str = '\n'.join([f'{k}: {v:.4f}' for k, v in res.items()])
        metric_str = metric_str + '\n'
        logger.info("====================================")
        logger.info("evaluate result:\n")
        logger.info(metric_str)
        logger.info("====================================")

        with open(output, 'w', encoding='utf-8') as fs:
          fs.write(f"metrics:\n{metric_str}\n")
      return main_metric
    return run_eval


  def get_feature_fn(self, max_seq_len=512, ext_params=None, rng=None, **kwargs):
    tokenizer = self.tokenizer
    mask_generator = NGramMaskGenerator(tokenizer)
    def example_to_feature(*example):
      '''
      sample: text, label
      return: ["input_ids", "input_mask", "token_type_id", "next_sentence_labels",
                  masked_lm_positions", "masked_lm_ids", "masked_lm_weights"]
      '''
      r = rng
      if r is None:
        r = random
      max_num_tokens = max_seq_len - 2
      segments = example[0].tolist().decode('utf-8')
      _tokens = ['[CLS]'] + tokenizer.tokenize(segments) + ['[SEP]']
      if mask_generator:
          tokens, lm_labels = mask_generator.mask_tokens(_tokens, r)
      token_ids = tokenizer.convert_tokens_to_ids(tokens)
      pad_id = tokenizer._convert_token_to_id(tokenizer.pad_token)
      seq_len = len(token_ids)
      pad_num = max_seq_len - seq_len
      token_ids = token_ids + [pad_id] * pad_num
      input_mask = [1] * seq_len + [0] * pad_num
      token_type_id = [0] * max_seq_len
      next_sentence_labels = -1
      masked_lm_positions = [i for i, lbl in enumerate(lm_labels) if lbl != 0]
      masked_lm_ids = [lbl for i, lbl in enumerate(lm_labels) if lbl != 0]
      if len(masked_lm_ids) < mask_generator.max_preds_per_seq:
        masked_lm_positions += [0] * (mask_generator.max_preds_per_seq - len(masked_lm_ids))
        masked_lm_ids += [-1] * (mask_generator.max_preds_per_seq - len(masked_lm_ids))
      masked_lm_weights = [1] * len(masked_lm_ids)
      features = (token_ids,
                  input_mask,
                  token_type_id,
                  next_sentence_labels,
                  masked_lm_positions,
                  masked_lm_ids,
                  masked_lm_weights)
      type_cast_op = transforms.c_transforms.TypeCast(mstype.int32)
      features = [np.array(f, dtype=np.int32) for f in features]
      return tuple(features)

    return example_to_feature

  def get_labels(self):
    """See base class."""
    return [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
  
  def load_txt_data(self, path, type=None, max_seq_len=512):
    examples = []
    with open(path, encoding='utf-8') as fs:
        current_chunk = ''
        current_length = 0
        for l in fs:
          l = l.replace('\t', '')
          if self.args.debug:
            if len(examples) >= 3000:
              return examples
          if type == 'JW':
            if random.random() > 0.5:
              continue
          if '\n' in l:
            l_list = l.split('\n')
          else:
            l_list = [l]
          for l in l_list:
            if len(l) == 0:
              continue
            current_chunk += l
            if len(current_chunk) >= max_seq_len:
              example = ExampleInstance(segments=[current_chunk[: max_seq_len - 2]])
              if len(example.segments[0]) < 510:
                print(len(example.segments[0]))
              # examples.append(np.array([current_chunk[: max_seq_len - 2]]))
              # example = np.array([current_chunk[: max_seq_len - 2]])
              examples.append(example)
              current_chunk = current_chunk[max_seq_len - 2: ]
    return ExampleSet(examples)

  def load_data(self, path, type=None, max_seq_len=512):
    import pickle
    examples = []
    with open(path, 'rb') as fs:
      data = pickle.load(fs)
    for cat, files in data.items():
      for path, lines in files.items():
        current_chunk = ''
        current_length = 0
        for l in lines:
          if self.args.debug:
            if len(examples) >= 30000:
              return ExampleSet(examples)
          # if '\n' in l:
          #   l_list = l.split('\n')
          # else:
          #   l_list = [l]
          # for l in l_list:
          if len(l) == 0:
            continue
          current_chunk += l
          if len(current_chunk) >= max_seq_len:
            if current_chunk[0] in ['O', '。', '？', '！', '，', '、', '；', '：', '|']:
              current_chunk = current_chunk[1:]
            example = ExampleInstance(segments=[current_chunk[: max_seq_len - 2]])
            if len(example.segments[0]) < 510:
              print(len(example.segments[0]))
            # examples.append(np.array([current_chunk[: max_seq_len - 2]]))
            examples.append(example)
            current_chunk = current_chunk[max_seq_len - 2: ]
    return ExampleSet(examples)


  def get_model(self):
    from MindsporeTrainer.modeling.models import BertPreTraining
    from MindsporeTrainer.modeling import build_transformer_model
    if self.args.fp16:
      compute_type = mstype.float16
    else:
      compute_type = mstype.float32
    model, config = build_transformer_model(self.args.model_config, 
                                            model=BertPreTraining, 
                                            compute_type=compute_type, 
                                            padding_idx=self.tokenizer._convert_token_to_id(self.tokenizer.pad_token))
    with open(os.path.join(self.args.output_dir, 'config.json'), 'w', encoding='utf-8') as f:
      f.write(config.to_json_string())
    copyfile(self.args.vocab_path, os.path.join(self.args.output_dir, 'vocab.txt'))
    return model
    # return partial_class

  def get_loss(self, *args, **kwargs):
    from MindsporeTrainer.modeling.loss import BertPretrainingLoss
    return BertPretrainingLoss(self.tokenizer.vocab_size)

  def get_eval_head(self, *args, **kwargs):
    from MindsporeTrainer.modeling.models import BertEvalHead
    return BertEvalHead(self.tokenizer.vocab_size)

  def get_opt_fn(self, *args, **kwargs):
    # def opt_fn(args, model, **kwargs):
    #   return Momentum(filter(lambda x: x.requires_grad, model.get_parameters()), 0.01, 0.9)
    return None





# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import random
import mmap
import numpy as np
from bisect import bisect
import loguru
import numpy as np
import mindspore.dataset as ds
import mindspore.dataset.transforms as C
import mindspore.common.dtype as mstype
logger=loguru.logger

def create_dynamic_dataset(examples, feature_fn, batch_size, output_columns, column_names=["example", "label"],
                           buffer_size=1000, repeat=1, num_workers=1, num_shards=None,
                           shard_id=None, type_cast_op=[C.c_transforms.TypeCast(mstype.int32)], shuffle=True):
    '''
        Provide dynamic data set.
        Args:
            examples: text example list
            feature_fn: function for generating mask etc.
    '''
    dataset = ds.GeneratorDataset(examples, 
                                  shuffle=shuffle,
                                  column_names=column_names,
                                  num_shards=num_shards,
                                  shard_id=shard_id,
                                  num_parallel_workers=num_workers
                                )
    # test_feature_fn(examples, feature_fn)
    for fn, name in feature_fn:
        dataset = dataset.map(fn,
                              input_columns=name,
                              output_columns=name,
                              column_order=output_columns,
                              num_parallel_workers=num_workers
                            )
    if type_cast_op is not None:
        for op, col in type_cast_op:
            dataset = dataset.map(operations=op, input_columns=col, num_parallel_workers=num_workers)
    dataset = dataset.batch(batch_size=batch_size, drop_remainder=True, num_parallel_workers=num_workers)
    # print_example(dataset)
    return dataset


def print_example(dataset, num_batch=1):
    iter = dataset.create_dict_iterator(output_numpy=True)
    for i in range(num_batch):
        e = next(iter)
        print(f'sample shape: {e["img"].shape}\nlabels: {e["label"]}')
    dataset.reset()

def test_feature_fn(example, feature_fn):
    e = example[0]
    for fn, name in feature_fn:
        e = fn(*e)
    print(e)